@@ -23,6 +23,8 @@
 #include <gtk/gtk.h>
 
 #include "htmlengine-edit-clueflowstyle.h"
+#include "htmlengine-edit-copy.h"
+#include "htmlengine-edit-cut.h"
 #include "htmlengine-edit-delete.h"
 #include "htmlengine-edit-fontstyle.h"
 #include "htmlengine-edit-insert.h"
@@ -33,6 +35,7 @@
 
 #include "gtkhtml-embedded.h"
 #include "gtkhtml-keybinding.h"
+#include "gtkhtml-stream.h"
 #include "gtkhtml-private.h"
 
 
@@ -57,6 +60,19 @@ enum {
 };
 static guint signals [LAST_SIGNAL] = { 0 };
 
+
+/* Values for selection information.  FIXME: what about COMPOUND_STRING and
+   TEXT?  */
+enum _TargetInfo {  
+	TARGET_STRING,
+	TARGET_TEXT,
+	TARGET_COMPOUND_TEXT
+};
+typedef enum _TargetInfo TargetInfo;
+
+/* Interval for scrolling during selection.  */
+#define SCROLL_TIMEOUT_INTERVAL 10
+
 
 static GtkHTMLParagraphStyle
 clueflow_style_to_paragraph_style (HTMLClueFlowStyle style)
@@ -210,7 +226,7 @@ idle_handler (gpointer data)
 	gtk_adjustment_set_value (GTK_LAYOUT (html)->hadjustment, (gfloat) engine->x_offset);
 	gtk_adjustment_set_value (GTK_LAYOUT (html)->vadjustment, (gfloat) engine->y_offset);
 
-	gtk_html_calc_scrollbars (html);
+	gtk_html_private_calc_scrollbars (html);
 
 	html_engine_flush_draw_queue (engine);
 
@@ -376,6 +392,112 @@ connect_adjustments (GtkHTML *html,
 		html->hadj_connection = 0;
 }
 
+
+/* Scroll timeout handling.  */
+
+static void
+inc_adjustment (GtkAdjustment *adj, gint doc_width, gint alloc_width, gint inc)
+{
+	gfloat value;
+	gint max;
+
+	value = adj->value + (gfloat) inc;
+	
+	if (doc_width > alloc_width)
+		max = doc_width - alloc_width;
+	else
+		max = 0;
+
+	if (value > (gfloat) max)
+		value = (gfloat) max;
+	else if (value < 0)
+		value = 0.0;
+
+	gtk_adjustment_set_value (adj, value);
+}
+
+static gint
+scroll_timeout_cb (gpointer data)
+{
+	GtkWidget *widget;
+	GtkHTML *html;
+	GtkLayout *layout;
+	gint x_scroll, y_scroll;
+	gint x, y;
+
+	GDK_THREADS_ENTER ();
+
+	widget = GTK_WIDGET (data);
+	html = GTK_HTML (data);
+
+	gdk_window_get_pointer (widget->window, &x, &y, NULL);
+
+	if (x < 0) {
+		x_scroll = x;
+		x = 0;
+	} else if (x >= widget->allocation.width) {
+		x_scroll = x - widget->allocation.width + 1;
+		x = widget->allocation.width;
+	} else {
+		x_scroll = 0;
+	}
+	x_scroll /= 2;
+
+	if (y < 0) {
+		y_scroll = y;
+		y = 0;
+	} else if (y >= widget->allocation.height) {
+		y_scroll = y - widget->allocation.height + 1;
+		y = widget->allocation.height;
+	} else {
+		y_scroll = 0;
+	}
+	y_scroll /= 2;
+
+	if (html->in_selection && (x_scroll != 0 || y_scroll != 0)) {
+		HTMLEngine *engine;
+
+		engine = html->engine;
+		html_engine_select_region (engine,
+					   html->selection_x1, html->selection_y1,
+					   x + engine->x_offset, y + engine->y_offset,
+					   TRUE);
+	}
+
+	layout = GTK_LAYOUT (widget);
+
+	inc_adjustment (layout->hadjustment, html_engine_get_doc_width (html->engine),
+			widget->allocation.width, x_scroll);
+	inc_adjustment (layout->vadjustment, html_engine_get_doc_height (html->engine),
+			widget->allocation.height, y_scroll);
+
+	GDK_THREADS_LEAVE ();
+
+	return TRUE;
+}
+
+static void
+setup_scroll_timeout (GtkHTML *html)
+{
+	if (html->scroll_timeout_id != 0)
+		return;
+
+	html->scroll_timeout_id = gtk_timeout_add (SCROLL_TIMEOUT_INTERVAL,
+						   scroll_timeout_cb, html);
+
+	scroll_timeout_cb (html);
+}
+
+static void
+remove_scroll_timeout (GtkHTML *html)
+{
+	if (html->scroll_timeout_id == 0)
+		return;
+
+	gtk_timeout_remove (html->scroll_timeout_id);
+	html->scroll_timeout_id = 0;
+}
+
 
 /* GtkObject methods.  */
 
@@ -397,6 +519,9 @@ destroy (GtkObject *object)
 	if (html->idle_handler_id != 0)
 		gtk_idle_remove (html->idle_handler_id);
 
+	if (html->scroll_timeout_id != 0)
+		gtk_timeout_remove (html->scroll_timeout_id);
+
 	gtk_object_destroy (GTK_OBJECT (html->engine));
 
 	if (GTK_OBJECT_CLASS (parent_class)->destroy != NULL)
@@ -453,10 +578,6 @@ realize (GtkWidget *widget)
 				| GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK
 				| GDK_KEY_PRESS_MASK | GDK_KEY_RELEASE_MASK));
 
-	/* FIXME */
-	html_settings_set_bgcolor (html->engine->settings, 
-				   &widget->style->bg[GTK_STATE_NORMAL]);
-
 	html_engine_realize (html->engine, html->layout.bin_window);
 
 	gdk_window_set_cursor (widget->window, html->arrow_cursor);
@@ -527,7 +648,7 @@ size_allocate (GtkWidget *widget, GtkAllocation *allocation)
 
 		html_engine_calc_size (html->engine);
 
-		gtk_html_calc_scrollbars (html);
+		gtk_html_private_calc_scrollbars (html);
 	}
 }
 
@@ -556,16 +677,16 @@ motion_notify_event (GtkWidget *widget,
 		x = event->x;
 		y = event->y;
 	}
-		
 
 	obj = html_engine_get_object_at (engine,
 					 x + engine->x_offset, y + engine->y_offset,
 					 NULL, FALSE);
 	if (html->button_pressed) {
 		if (obj) {
-
 			type = HTML_OBJECT_TYPE (obj);
 
+			/* FIXME this is broken */
+
 			if (type == HTML_TYPE_BUTTON ||
 			    type ==  HTML_TYPE_CHECKBOX ||
 			    type ==  HTML_TYPE_EMBEDDED ||
@@ -575,12 +696,23 @@ motion_notify_event (GtkWidget *widget,
 			    type ==  HTML_TYPE_SELECT ||
 			    type ==  HTML_TYPE_TEXTAREA ||
 			    type ==  HTML_TYPE_TEXTINPUT ) {
-
 				return FALSE;
 			}
 		}
+
 		html->in_selection = TRUE;
-		
+
+		if (x < 0 || x >= widget->allocation.width
+		    || y < 0 || y >= widget->allocation.height)
+			setup_scroll_timeout (html);
+		else
+			remove_scroll_timeout (html);
+
+		/* This will put the mark at the position of the
+                   previous click.  */
+		if (engine->mark == NULL && engine->editable)
+			html_engine_set_mark (engine);
+
 		html_engine_select_region (engine,
 					   html->selection_x1, html->selection_y1,
 					   x + engine->x_offset, y + engine->y_offset,
@@ -590,9 +722,9 @@ motion_notify_event (GtkWidget *widget,
 			html_engine_jump_at (engine,
 					     event->x + engine->x_offset,
 					     event->y + engine->y_offset);
-		
 		return TRUE;
 	}
+
 	if (obj != NULL)
 		url = html_object_get_url (obj);
 	else
@@ -631,12 +763,48 @@ button_press_event (GtkWidget *widget,
 {
 	GtkHTML *html;
 	HTMLEngine *engine;
+	gint value;
 
 	html = GTK_HTML (widget);
 	engine = html->engine;
 
 	gtk_widget_grab_focus (widget);
 
+	if (event->type == GDK_BUTTON_PRESS) {
+		GtkAdjustment *vadj;
+			
+		vadj = GTK_LAYOUT (widget)->vadjustment;
+		
+		switch (event->button) {
+		case 4:
+			/* Mouse wheel scroll up.  */
+			value = vadj->value - vadj->step_increment * 3;
+			
+			if (value < vadj->lower)
+				value = vadj->lower;
+			
+			gtk_adjustment_set_value (vadj, value);
+			return TRUE;
+			break;
+		case 5:
+			/* Mouse wheel scroll down.  */
+			value = vadj->value + vadj->step_increment * 3;
+			
+			if (value > (vadj->upper - vadj->page_size))
+				value = vadj->upper - vadj->page_size;
+			
+			gtk_adjustment_set_value (vadj, value);
+			return TRUE;
+			break;
+		case 2:
+			gtk_html_request_paste (widget, event->time);
+			return TRUE;
+			break;
+		default:
+			break;
+		}
+	}
+
 	if (html_engine_get_editable (engine)) {
 		html_engine_jump_at (engine,
 				     event->x + engine->x_offset,
@@ -658,7 +826,7 @@ button_press_event (GtkWidget *widget,
 
 	html->button_pressed = TRUE;
 
-	html_engine_unselect_all (engine, TRUE);
+	html_engine_disable_selection (engine);
 
 	return TRUE;
 }
@@ -680,10 +848,13 @@ button_release_event (GtkWidget *widget,
 	html->button_pressed = FALSE;
 
 	if (html->in_selection) {
+		gtk_selection_owner_set (widget, GDK_SELECTION_PRIMARY, event->time);
 		html->in_selection = FALSE;
 		update_styles (html);
 	}
 
+	remove_scroll_timeout (html);
+
 	return TRUE;
 }
 
@@ -709,6 +880,134 @@ focus_out_event (GtkWidget *widget,
 	return FALSE;
 }
 
+
+/* X11 selection support.  */
+
+static void
+selection_get (GtkWidget        *widget, 
+	       GtkSelectionData *selection_data_ptr,
+	       guint             info,
+	       guint             time)
+{
+	GtkHTML *html;
+	gchar *selection_string;
+	
+	g_return_if_fail (widget != NULL);
+	g_return_if_fail (GTK_IS_HTML (widget));
+	
+	html = GTK_HTML (widget);
+	selection_string = html_engine_get_selection_string (html->engine);
+	
+	if (selection_string != NULL) {
+		if (info == TARGET_STRING)
+			{
+				gtk_selection_data_set (selection_data_ptr,
+							GDK_SELECTION_TYPE_STRING, 8,
+							selection_string, 
+							strlen (selection_string));
+			}
+		else if ((info == TARGET_TEXT) || (info == TARGET_COMPOUND_TEXT))
+			{
+				guchar *text;
+				GdkAtom encoding;
+				gint format;
+				gint new_length;
+				
+				gdk_string_to_compound_text (selection_string, 
+							     &encoding, &format,
+							     &text, &new_length);
+
+				gtk_selection_data_set (selection_data_ptr,
+							encoding, format,
+							text, new_length);
+				gdk_free_compound_text (text);
+			}
+		g_free (selection_string);
+	}
+}
+
+/* receive a selection */
+/* Signal handler called when the selections owner returns the data */
+static void
+selection_received (GtkWidget *widget,
+		    GtkSelectionData *selection_data, 
+		    guint time)
+{
+	g_return_if_fail (widget != NULL);
+	g_return_if_fail (GTK_IS_HTML (widget));
+	g_return_if_fail (selection_data != NULL);
+	
+	printf("got selection from system\n");
+	
+#if 0
+	/* **** IMPORTANT **** Check to see if retrieval succeeded  */
+	/* If we have more selection types we can ask for, try the next one,
+	   until there are none left */
+	if (selection_data->length < 0) {
+		struct _zvtprivate *zp = _ZVT_PRIVATE(widget);
+		
+		/* now, try again with next selection type */
+		if (gtk_html_request_paste(widget, zp->lastselectiontype+1, time)==0)
+			g_print ("Selection retrieval failed\n");
+		return;
+	}
+#endif
+	
+	/* we will get a selection type of atom(UTF-8) for utf text,
+	   perhaps that needs to do something different if the terminal
+	   isn't actually in utf8 mode? */
+	
+	/* Make sure we got the data in the expected form */
+	if (selection_data->type != GDK_SELECTION_TYPE_STRING) {
+		g_print ("Selection \"STRING\" was not returned as strings!\n");
+		return;
+	}
+	
+	if (selection_data->length) {
+		printf ("selection text \"%.*s\"\n",
+			selection_data->length, selection_data->data); 
+
+		html_engine_disable_selection (GTK_HTML (widget)->engine);
+		html_engine_insert (GTK_HTML (widget)->engine, 
+				    selection_data->data,
+				    selection_data->length);
+	}
+}  
+
+int
+gtk_html_request_paste (GtkWidget *widget, gint32 time)
+{
+  GdkAtom string_atom;
+
+  string_atom = gdk_atom_intern ("STRING", FALSE);
+
+  if (string_atom == GDK_NONE) {
+    g_warning("WARNING: Could not get string atom\n");
+  }
+  /* And request the "STRING" target for the primary selection */
+    gtk_selection_convert (widget, GDK_SELECTION_PRIMARY, string_atom,
+			   time);
+  return 1;
+}
+
+
+static gint
+selection_clear_event (GtkWidget *widget,
+		       GdkEventSelection *event)
+{
+	GtkHTML *html;
+
+	if (! gtk_selection_clear (widget, event))
+		return FALSE;
+
+	html = GTK_HTML (widget);
+
+	html_engine_disable_selection (html->engine);
+	html->in_selection = FALSE;
+
+	return TRUE;
+}
+
 
 static void
 set_adjustments (GtkLayout     *layout,
@@ -882,6 +1181,9 @@ class_init (GtkHTMLClass *klass)
 	widget_class->button_release_event = button_release_event;
 	widget_class->focus_in_event = focus_in_event;
 	widget_class->focus_out_event = focus_out_event;
+	widget_class->selection_get = selection_get;
+	widget_class->selection_received = selection_received;
+	widget_class->selection_clear_event = selection_clear_event;
 
 	layout_class->set_scroll_adjustments = set_adjustments;
 }
@@ -889,6 +1191,13 @@ class_init (GtkHTMLClass *klass)
 static void
 init (GtkHTML* html)
 {
+	static const GtkTargetEntry targets[] = {
+		{ "STRING", 0, TARGET_STRING },
+		{ "TEXT",   0, TARGET_TEXT }, 
+		{ "COMPOUND_TEXT", 0, TARGET_COMPOUND_TEXT },
+	};
+	static const gint n_targets = sizeof(targets) / sizeof(targets[0]);
+
 	GTK_WIDGET_SET_FLAGS (GTK_WIDGET (html), GTK_CAN_FOCUS);
 	GTK_WIDGET_SET_FLAGS (GTK_WIDGET (html), GTK_APP_PAINTABLE);
 
@@ -911,12 +1220,17 @@ init (GtkHTML* html)
 	html->load_in_progress = TRUE;
 
 	html->idle_handler_id = 0;
+	html->scroll_timeout_id = 0;
 
 	html->paragraph_style = GTK_HTML_PARAGRAPH_STYLE_NORMAL;
 	html->paragraph_alignment = GTK_HTML_PARAGRAPH_ALIGNMENT_LEFT;
 	html->paragraph_indentation = 0;
 
 	html->insertion_font_style = GTK_HTML_FONT_STYLE_DEFAULT;
+
+	gtk_selection_add_targets (GTK_WIDGET (html),
+				   GDK_SELECTION_PRIMARY,
+				   targets, n_targets);
 }
 
 
@@ -946,10 +1260,23 @@ gtk_html_get_type (void)
 GtkWidget *
 gtk_html_new (void)
 {
-	GtkHTML *html;
+	GtkWidget *html;
 
 	html = gtk_type_new (gtk_html_get_type ());
-	
+	gtk_html_construct (html);
+	return html;
+}
+
+void
+gtk_html_construct (GtkWidget *htmlw)
+{
+	GtkHTML *html;
+
+	g_return_if_fail (htmlw != NULL);
+	g_return_if_fail (GTK_IS_HTML (htmlw));
+
+	html = GTK_HTML (htmlw);
+
 	html->engine = html_engine_new ();
 	html->engine->widget = html; /* FIXME FIXME */
 
@@ -971,8 +1298,6 @@ gtk_html_new (void)
 			    GTK_SIGNAL_FUNC (html_engine_submit_cb), html);
 	gtk_signal_connect (GTK_OBJECT (html->engine), "object_requested",
 			    GTK_SIGNAL_FUNC (html_engine_object_requested_cb), html);
-
-	return GTK_WIDGET (html);
 }
 
 
@@ -998,12 +1323,6 @@ gtk_html_allow_selection (GtkHTML *html,
 }
 
 
-void
-gtk_html_parse (GtkHTML *html)
-{
-	g_warning ("Calling %s() is now DEPRECATED and not required anymore.", __FUNCTION__);
-}
-
 GtkHTMLStreamHandle
 gtk_html_begin (GtkHTML *html, const gchar *url)
 {
@@ -1034,7 +1353,7 @@ gtk_html_end (GtkHTML *html,
 	      GtkHTMLStreamHandle handle,
 	      GtkHTMLStreamStatus status)
 {
-	gtk_html_stream_end (handle, status);
+	gtk_html_stream_close (handle, status);
 
 	html->load_in_progress = FALSE;
 }
@@ -1046,9 +1365,23 @@ gtk_html_get_title (GtkHTML *html)
 	g_return_val_if_fail (html != NULL, NULL);
 	g_return_val_if_fail (GTK_IS_HTML (html), NULL);
 
+	if (html->engine->title == NULL)
+		return NULL;
+
 	return html->engine->title->str;
 }
 
+
+gboolean
+gtk_html_jump_to_anchor (GtkHTML *html,
+			 const gchar *anchor)
+{
+	g_return_val_if_fail (html != NULL, FALSE);
+	g_return_val_if_fail (GTK_IS_HTML (html), FALSE);
+
+	return html_engine_goto_anchor (html->engine, anchor);
+}
+
 
 gboolean
 gtk_html_save (GtkHTML *html,
@@ -1064,25 +1397,33 @@ gtk_html_save (GtkHTML *html,
 
 
 void
-gtk_html_calc_scrollbars (GtkHTML *html)
+gtk_html_private_calc_scrollbars (GtkHTML *html)
 {
+	GtkLayout *layout;
+	GtkAdjustment *vadj, *hadj;
 	gint width, height;
 
 	height = html_engine_get_doc_height (html->engine);
 	width = html_engine_get_doc_width (html->engine);
 
-	GTK_LAYOUT (html)->vadjustment->lower = 0;
-	GTK_LAYOUT (html)->vadjustment->page_size = html->engine->height;
-	GTK_LAYOUT (html)->vadjustment->step_increment = 14; /* FIXME */
-	GTK_LAYOUT (html)->vadjustment->page_increment = html->engine->height;
+	layout = GTK_LAYOUT (html);
+	hadj = layout->hadjustment;
+	vadj = layout->vadjustment;
 
-	GTK_LAYOUT (html)->hadjustment->lower = 0;
-	GTK_LAYOUT (html)->hadjustment->page_size = html->engine->width;
-	GTK_LAYOUT (html)->hadjustment->step_increment = 14; /* FIXME */
-	GTK_LAYOUT (html)->hadjustment->page_increment = html->engine->width;
+	vadj->lower = 0;
+	vadj->upper = height;
+	vadj->page_size = html->engine->height;
+	vadj->step_increment = 14; /* FIXME */
+	vadj->page_increment = html->engine->height;
 
-	if (width != GTK_LAYOUT (html)->width || height != GTK_LAYOUT (html)->height)
-		gtk_layout_set_size (GTK_LAYOUT (html), width, height);
+	hadj->lower = 0.0;
+	hadj->upper = width;
+	hadj->page_size = html->engine->width;
+	hadj->step_increment = 14; /* FIXME */
+	hadj->page_increment = html->engine->width;
+
+	if (width != layout->width || height != layout->height)
+		gtk_layout_set_size (layout, width, height);
 }
 
 
@@ -1206,7 +1547,7 @@ gtk_html_cut (GtkHTML *html)
 	g_return_if_fail (html != NULL);
 	g_return_if_fail (GTK_IS_HTML (html));
 
-	html_engine_cut (html->engine);
+	html_engine_cut (html->engine, TRUE);
 }
 
 void
@@ -1224,7 +1565,7 @@ gtk_html_paste (GtkHTML *html)
 	g_return_if_fail (html != NULL);
 	g_return_if_fail (GTK_IS_HTML (html));
 
-	html_engine_paste (html->engine);
+	html_engine_paste (html->engine, TRUE);
 }
 
 
@@ -1248,3 +1589,10 @@ gtk_html_redo (GtkHTML *html)
 	html_engine_redo (html->engine);
 }
 
+/* misc utils */
+
+void
+gtk_html_set_default_background_color (GtkHTML *html, GdkColor *c)
+{
+	html_settings_set_color (html->engine->defaultSettings, HTMLBgColor, c);
+}
