@@ -22,6 +22,7 @@
 */
 #include "htmlclue.h"
 #include "htmlcluev.h"
+#include "htmlsearch.h"
 
 
 #define HC_CLASS(x) (HTML_CLUE_CLASS (HTML_OBJECT (x)->klass))
@@ -35,17 +36,12 @@ HTMLObjectClass *parent_class = NULL;
 static void
 destroy (HTMLObject *o)
 {
-	HTMLObject *obj;
+	HTMLObject *p;
 
-	while (HTML_CLUE (o)->head) {
-		HTML_CLUE (o)->curr = HTML_CLUE (o)->head->next;
-		obj = HTML_CLUE(o)->head;
-		HTML_CLUE (o)->head = HTML_CLUE (o)->curr;
+	for (p = HTML_CLUE (o)->head; p != NULL; p = p->next)
+		html_object_destroy (p);
 
-		html_object_destroy (obj);
-	}
-
-	HTML_OBJECT_CLASS (&html_object_class)->destroy (o);
+	HTML_OBJECT_CLASS (parent_class)->destroy (o);
 }
 
 static void
@@ -145,6 +141,8 @@ reset (HTMLObject *clue)
 		html_object_reset (obj);
 
 	HTML_CLUE (clue)->curr = NULL;
+
+	(* HTML_OBJECT_CLASS (parent_class)->reset) (HTML_OBJECT (clue));
 }
 
 static gboolean
@@ -181,9 +179,6 @@ calc_preferred_width (HTMLObject *o,
 	gint prefWidth = 0;
 	HTMLObject *obj;
 	
-	if (o->flags & HTML_OBJECT_FLAG_FIXEDWIDTH)
-		return o->width;
-
 	for (obj = HTML_CLUE (o)->head; obj != 0; obj = obj->next) {
 		gint w;
 
@@ -195,6 +190,7 @@ calc_preferred_width (HTMLObject *o,
 	return prefWidth;
 }
 
+/* FIXME: This should be in HTMLClueV.  */
 static gint
 calc_min_width (HTMLObject *o,
 		HTMLPainter *painter)
@@ -210,11 +206,6 @@ calc_min_width (HTMLObject *o,
 			minWidth = w;
 	}
 	
-	if ((o->flags & HTML_OBJECT_FLAG_FIXEDWIDTH)) {
-		if (o->width > minWidth)
-			minWidth = o->width;
-	}
-	
 	return minWidth;
 }
 
@@ -345,6 +336,40 @@ appended (HTMLClue *clue, HTMLClue *aclue)
 	return FALSE;
 }
 
+static gboolean
+search (HTMLObject *obj, HTMLSearch *info)
+{
+	HTMLObject *cur;
+	HTMLClue *clue = HTML_CLUE (obj);
+	gboolean next = FALSE;
+
+	printf ("search clue\n");
+
+	/* search_next? */
+	if (html_search_child_on_stack (info, obj)) {
+		cur  = html_search_pop (info);
+		cur = (info->forward) ? cur->next : cur->prev;
+		next = TRUE;
+	} else {
+		cur = (info->forward) ? clue->head : clue->tail;
+	}
+
+	while (cur) {
+		html_search_push (info, cur);
+		if (html_object_search (cur, info)) {
+			return TRUE;
+		}
+		html_search_pop (info);
+		cur = (info->forward) ? cur->next : cur->prev;
+	}
+
+	if (next) {
+		return html_search_next_parent (info);
+	}
+
+	return FALSE;
+}
+
 
 void
 html_clue_type_init (void)
@@ -380,6 +405,7 @@ html_clue_class_init (HTMLClueClass *klass,
 	object_class->forall = forall;
 	object_class->is_container = is_container;
 	object_class->save = save;
+	object_class->search = search;
 
 	/* HTMLClue methods.  */
 	klass->get_left_clear = get_left_clear;
@@ -434,12 +460,22 @@ html_clue_find_free_area (HTMLClue *clue, gint y,
 void
 html_clue_append_right_aligned (HTMLClue *clue, HTMLClue *aclue)
 {
+	g_assert (clue != NULL);
+	g_assert (aclue != NULL);
+
+	html_object_change_set (HTML_OBJECT (clue), HTML_OBJECT (aclue)->change);
+
 	(* HC_CLASS (clue)->append_right_aligned) (clue, aclue);
 }
 
 void
 html_clue_append_left_aligned (HTMLClue *clue, HTMLClue *aclue)
 {
+	g_assert (clue != NULL);
+	g_assert (aclue != NULL);
+
+	html_object_change_set (HTML_OBJECT (clue), HTML_OBJECT (aclue)->change);
+
 	(* HC_CLASS (clue)->append_left_aligned) (clue, aclue);
 }
 
@@ -496,6 +532,8 @@ html_clue_append_after (HTMLClue *clue,
 	g_return_if_fail (o != NULL);
 	g_return_if_fail (where != NULL);
 
+	html_object_change_set (HTML_OBJECT (clue), o->change);
+
 	tail = get_tail (o);
 
 	if (where->next != NULL)
@@ -527,6 +565,8 @@ html_clue_append (HTMLClue *clue,
 	g_return_if_fail (clue != NULL);
 	g_return_if_fail (o != NULL);
 
+	html_object_change_set (HTML_OBJECT (clue), o->change);
+
 	tail = get_tail (o);
 
 	if (! clue->head) {
@@ -561,6 +601,8 @@ html_clue_prepend (HTMLClue *clue,
 	g_return_if_fail (clue != NULL);
 	g_return_if_fail (o != NULL);
 
+	html_object_change_set (HTML_OBJECT (clue), o->change);
+
 	tail = get_tail (o);
 
 	if (! clue->head) {
