@@ -25,6 +25,8 @@
 
 /* WARNING: it must always be the child of a clue.  */
 
+#include <config.h>
+#include <string.h>
 #include "htmlclue.h"
 #include "htmlclueflow.h"
 #include "htmlcluealigned.h"
@@ -32,6 +34,8 @@
 #include "htmlvspace.h"
 #include "htmllinktextmaster.h"
 #include "htmltextslave.h"	/* FIXME */
+#include "htmlsearch.h"
+#include "htmlentity.h"
 
 
 HTMLClueFlowClass html_clueflow_class;
@@ -172,7 +176,7 @@ add_post_padding (HTMLClueFlow *flow,
 		    && ! is_header (next))
 			goto add_pad;
 
-		if (is_header (flow))
+		if (is_header (flow) &&  ! is_header (next))
 			goto add_pad;
 
 		return;
@@ -226,48 +230,52 @@ static gboolean
 should_break (HTMLObject *obj)
 {
 	/* end: last char of obj->prev, beg: beggining char of obj */
-	guchar end=0, beg=0;
-	HTMLObject *prev;
+	guchar end=' ', beg=' ';
+	HTMLObject *next;
 
 	if (!obj)
 		return TRUE;
-	prev = obj->prev;
+	next = obj->next;
 
 	/* it is on the beginning of Clue */
-	if (!prev)
-		return FALSE;
+	if (!next)
+		return TRUE;
 
 	/* skip prev Master */
-	if ((HTML_OBJECT_TYPE (prev) == HTML_TYPE_TEXTMASTER ||
-	     HTML_OBJECT_TYPE (prev) == HTML_TYPE_LINKTEXTMASTER) &&
-	    HTML_OBJECT_TYPE (obj)  == HTML_TYPE_TEXTSLAVE  &&
-	    HTML_TEXT_SLAVE (obj)->owner == HTML_TEXT_MASTER (prev)) {
-		prev = prev->prev;
-		/* it is Master->Slave on the beginning */
-		if (!prev)
-			return FALSE;
+	if ((HTML_OBJECT_TYPE (obj) == HTML_TYPE_TEXTMASTER ||
+	     HTML_OBJECT_TYPE (obj) == HTML_TYPE_LINKTEXTMASTER) &&
+	    HTML_OBJECT_TYPE (next)  == HTML_TYPE_TEXTSLAVE  &&
+	    HTML_TEXT_SLAVE (next)->owner == HTML_TEXT_MASTER (obj)) {
+		next = next->next;
+		/* it is Master->Slave */
+		if (!next)
+			return TRUE;
 	}
 
 	/* get end */
-	if (HTML_OBJECT_TYPE (prev) == HTML_TYPE_TEXTSLAVE) {
+	if (HTML_OBJECT_TYPE (obj) == HTML_TYPE_TEXTSLAVE) {
 
-		HTMLTextSlave *slave = HTML_TEXT_SLAVE (prev);
+		HTMLTextSlave *slave = HTML_TEXT_SLAVE (obj);
 		end = HTML_TEXT (slave->owner)->text [slave->posStart + slave->posLen - 1];
 	} else
-		if (HTML_OBJECT_TYPE (prev) == HTML_TYPE_TEXTMASTER ||
-		    HTML_OBJECT_TYPE (prev) == HTML_TYPE_LINKTEXTMASTER) {
+		if (HTML_OBJECT_TYPE (obj) == HTML_TYPE_TEXTMASTER ||
+		    HTML_OBJECT_TYPE (obj) == HTML_TYPE_LINKTEXTMASTER) {
 
-			HTMLText *master = HTML_TEXT (prev);
-			end = master->text [master->text_len - 1];
+			HTMLText *master = HTML_TEXT (obj);
+			if (master->text_len > 0)
+				end = master->text [master->text_len - 1];
+			else
+				/* XXX is this the correct return value? */
+				end = ' ';
 		}
 
 	/* get beg */
-	if (HTML_OBJECT_TYPE (obj) == HTML_TYPE_TEXTSLAVE) {
-		HTMLTextSlave *slave = HTML_TEXT_SLAVE (obj);
+	if (HTML_OBJECT_TYPE (next) == HTML_TYPE_TEXTSLAVE) {
+		HTMLTextSlave *slave = HTML_TEXT_SLAVE (next);
 		beg = HTML_TEXT (slave->owner)->text [slave->posStart];
-	} else if (HTML_OBJECT_TYPE (obj) == HTML_TYPE_TEXTMASTER ||
-		 HTML_OBJECT_TYPE (obj) == HTML_TYPE_LINKTEXTMASTER)
-		beg = HTML_TEXT (obj)->text [0];
+	} else if (HTML_OBJECT_TYPE (next) == HTML_TYPE_TEXTMASTER ||
+		   HTML_OBJECT_TYPE (next) == HTML_TYPE_LINKTEXTMASTER)
+		beg = HTML_TEXT (next)->text [0];
 
 	/* printf ("end: %c beg: %c ret: %d\n", end, beg, end == ' ' || beg == ' '); */
 
@@ -342,6 +350,7 @@ calc_size (HTMLObject *o,
 	gboolean changed;
 	gint old_ascent, old_descent, old_width;
 	gint runWidth = 0;
+	gboolean have_valign_top;
 
 	changed = FALSE;
 	old_ascent = o->ascent;
@@ -378,6 +387,8 @@ calc_size (HTMLObject *o,
 	newLine = FALSE;
 	firstLine = TRUE;
 
+	have_valign_top = FALSE;
+
 	while (obj != NULL) {
 		if (obj->flags & HTML_OBJECT_FLAG_NEWLINE) {
 			if (!a)
@@ -464,6 +475,7 @@ calc_size (HTMLObject *o,
 				&& ! (run->flags & HTML_OBJECT_FLAG_NEWLINE)
 				&& ! (run->flags & HTML_OBJECT_FLAG_ALIGNED)) {
 				HTMLFitType fit;
+				HTMLVAlignType valign;
 				gint width_left;
 
 				run->max_width = rmargin - lmargin;
@@ -490,10 +502,42 @@ calc_size (HTMLObject *o,
 				html_object_calc_size (run, painter);
 				runWidth += run->width;
 
-				if (run->ascent > a)
-					a = run->ascent;
-				if (run->descent > d)
-					d = run->descent;
+				valign = html_object_get_valign (run);
+
+				/* Algorithm for dealing vertical alignment.
+				   Elements with `HTML_VALIGN_BOTTOM' and
+				   `HTML_VALIGN_CENTER' can be handled
+				   immediately.	 Objects with
+				   `HTML_VALIGN_TOP', instead, need to know the
+				   total height of the line, so need to be
+				   handled last.  */
+
+				switch (valign) {
+				case HTML_VALIGN_CENTER: {
+					gint height;
+					gint half_height;
+
+					height = run->ascent + run->descent;
+					half_height = height / 2;
+					if (half_height > a)
+						a = half_height;
+					if (height - half_height > d)
+						d = height - half_height;
+					break;
+				}
+				case HTML_VALIGN_BOTTOM:
+					if (run->ascent > a)
+						a = run->ascent;
+					if (run->descent > d)
+						d = run->descent;
+					break;
+				case HTML_VALIGN_TOP:
+					have_valign_top = TRUE;
+					/* Do nothing for now.	*/
+					break;
+				default:
+					g_assert_not_reached ();
+				}
 
 				run = run->next;
 
@@ -504,7 +548,6 @@ calc_size (HTMLObject *o,
 
 				if (! is_pre && runWidth > rmargin - lmargin)
 					break;
-
 			}
 
 			/* If these objects do not fit in the current line and we are
@@ -588,9 +631,6 @@ calc_size (HTMLObject *o,
 
 			extra = 0;
 
-			o->ascent += a + d;
-			o->y += a + d;
-
 			if (w > o->width)
 				o->width = w;
 
@@ -604,18 +644,59 @@ calc_size (HTMLObject *o,
 				if (extra < 0)
 					extra = 0;
 			}
-			
-			while (line != obj) {
-				if (!(line->flags & HTML_OBJECT_FLAG_ALIGNED)) {
+
+			o->ascent += a + d;
+			o->y += a + d;
+
+			/* Update the height for `HTML_VALIGN_TOP' objects.  */
+			if (have_valign_top) {
+				HTMLObject *p;
+
+				for (p = line; p != obj; p = p->next) {
+					gint height, rest;
+
+					if (html_object_get_valign (p) != HTML_VALIGN_TOP)
+						continue;
+
+					height = p->ascent + p->descent;
+					rest = height - a;
+
+					if (rest > d) {
+						o->ascent += rest - d;
+						o->y += rest - d;
+						d = rest;
+					}
+				}
+
+				have_valign_top = FALSE;
+			}
+
+			for (; line != obj; line = line->next) {
+				if (line->flags & HTML_OBJECT_FLAG_ALIGNED)
+					continue;
+
+				/* FIXME max_ascent/max_descent -- not quite
+				   sure what they are for. */
+
+				switch (html_object_get_valign (line)) {
+				case HTML_VALIGN_BOTTOM:
 					line->y = o->ascent - d;
 					html_object_set_max_ascent (line, painter, a);
 					html_object_set_max_descent (line, painter, d);
-					if (clue->halign == HTML_HALIGN_CENTER
-					    || clue->halign == HTML_HALIGN_RIGHT) {
-						line->x += extra;
-					}
+					break;
+				case HTML_VALIGN_CENTER:
+					line->y = o->ascent - a - d + line->ascent;
+					break;
+				case HTML_VALIGN_TOP:
+					line->y = o->ascent - a - d + line->ascent;
+					break;
+				default:
+					g_assert_not_reached ();
 				}
-				line = line->next;
+
+				if (clue->halign == HTML_HALIGN_CENTER
+				    || clue->halign == HTML_HALIGN_RIGHT)
+					line->x += extra;
 			}
 
 			oldy = o->y;
@@ -801,6 +882,17 @@ check_point (HTMLObject *self,
 	return NULL;
 }
 
+static void
+append_selection_string (HTMLObject *self,
+			 GString *buffer)
+{
+	if (! self->selected)
+		return;
+
+	putchar ('\n');
+	g_string_append_c (buffer, '\n');
+}
+
 
 /* Saving support.  */
 
@@ -883,7 +975,7 @@ get_tag_for_style (const HTMLClueFlow *flow)
 		return "H6";
 	case HTML_CLUEFLOW_STYLE_ADDRESS:
 		return "ADDRESS";
-	case HTML_CLUEFLOW_STYLE_PRE:
+	case HTML_CLUEFLOW_STYLE_PRE:	
 		return "PRE";
 	case HTML_CLUEFLOW_STYLE_ITEMDOTTED:
 	case HTML_CLUEFLOW_STYLE_ITEMROMAN:
@@ -909,7 +1001,19 @@ halign_to_string (HTMLHAlignType halign)
 		return "LEFT";
 	}
 }
+
+static gboolean 
+is_similar (HTMLObject *self, HTMLObject *friend)
+{
 	
+	if (friend &&  HTML_OBJECT_TYPE (friend) == HTML_TYPE_CLUEFLOW) {
+		if ((HTML_CLUEFLOW (friend)->style == HTML_CLUEFLOW (self)->style)
+		    && (HTML_CLUEFLOW (friend)->level == HTML_CLUEFLOW (self)->level))
+			return TRUE;
+	}
+	return FALSE;
+}
+
 static gboolean
 save (HTMLObject *self,
       HTMLEngineSaveState *state)
@@ -917,15 +1021,33 @@ save (HTMLObject *self,
 	HTMLClueFlow *clueflow;
 	HTMLHAlignType halign;
 	const gchar *tag;
+	gboolean start = TRUE, end = TRUE;
 
 	clueflow = HTML_CLUEFLOW (self);
 	halign = HTML_CLUE (self)->halign;
 
+	if (self->prev != NULL
+	    && HTML_OBJECT_TYPE (HTML_CLUE (self)->tail) != HTML_TYPE_RULE
+	    && HTML_CLUEFLOW (self)->style != HTML_CLUEFLOW_STYLE_PRE) {
+		/* This is a nasty hack: the rule takes all of the space, so we
+                   don't want it to create a new newline if it's the last
+                   element on the paragraph.  Also, everything but the "normal"
+                   paragraph style would get an extra newline if we add a
+                   `<BR>'.  */
+		html_engine_save_output_string (state, "<BR>\n");
+	}
+
 	if (! write_indentation_tags (clueflow, state))
 		return FALSE;
 
 	tag = get_tag_for_style (clueflow);
 
+	if (is_similar (self, self->prev))
+		start = FALSE;
+
+	if (is_similar (self, self->next))
+		end = FALSE;
+	
 	/* Alignment tag.  */
 	if (halign != HTML_HALIGN_NONE && halign != HTML_HALIGN_LEFT) {
 		if (! html_engine_save_output_string (state, "<DIV ALIGN=")
@@ -935,7 +1057,7 @@ save (HTMLObject *self,
 	}
 
 	/* Start tag.  */
-	if (tag != NULL
+	if (tag != NULL && start
 	    && (! html_engine_save_output_string (state, "<")
 		|| ! html_engine_save_output_string (state, tag)
 		|| ! html_engine_save_output_string (state, ">")))
@@ -946,7 +1068,7 @@ save (HTMLObject *self,
 		return FALSE;
 
 	/* End tag.  */
-	if (tag != NULL) {
+	if (tag && end) {
 		if (! html_engine_save_output_string (state, "</")
 		    || ! html_engine_save_output_string (state, tag)
 		    || ! html_engine_save_output_string (state, ">"))
@@ -962,15 +1084,6 @@ save (HTMLObject *self,
 			return FALSE;
 	}
 
-	if (HTML_OBJECT_TYPE (HTML_CLUE (self)->tail) != HTML_TYPE_RULE
-	    && ! is_header (HTML_CLUEFLOW (self))) {
-		/* This comparison is a nasty hack: the rule takes all of the
-                   space, so we don't want it to create a new newline if it's
-                   the last element on the paragraph.  Also, headers would get
-                   extra space if we add a BR.  */
-		html_engine_save_output_string (state, "<BR>\n");
-	}
-
 	return TRUE;
 }
 
@@ -1037,6 +1150,254 @@ get_default_font_style (const HTMLClueFlow *self)
 	}
 }
 
+static void
+search_set_info (HTMLObject *cur, HTMLSearch *info, guint pos, guint len)
+{
+	guint text_len = 0;
+	guint cur_len;
+
+	printf ("se_info pos: %d len: %d\n", pos, len);
+
+	info->found_len = len;
+
+	if (info->found) {
+		g_list_free (info->found);
+		info->found = NULL;
+	}
+
+	while (cur) {
+		if (HTML_OBJECT_TYPE (cur) == HTML_TYPE_TEXTMASTER
+		    || HTML_OBJECT_TYPE (cur) == HTML_TYPE_LINKTEXTMASTER) {
+			cur_len = HTML_TEXT (cur)->text_len;
+			if (text_len + cur_len > pos) {
+				if (!info->found) {
+					info->start_pos = pos - text_len;
+				}
+				info->found = g_list_append (info->found, cur);
+			}
+			text_len += cur_len;
+			if (text_len >= pos+info->found_len) {
+				info->stop_pos = cur_len - (text_len - pos - info->found_len);
+				info->last     = HTML_OBJECT (cur);
+				return;
+			}
+		} else if (HTML_OBJECT_TYPE (cur) != HTML_TYPE_TEXTSLAVE) {
+			break;
+		}		
+		cur = cur->next;
+	}
+
+	g_assert_not_reached ();
+}
+
+/* search text objects ([TextMaster, LinkTextMaster], TextSlave*) */
+static gboolean
+search_text (HTMLObject **beg, HTMLSearch *info)
+{
+	HTMLObject *cur = *beg;
+	HTMLObject *end = cur;
+	HTMLObject *head;
+	guchar *par, *pp;
+	guint text_len;
+	guint eq_len;
+	gint  pos;
+	gboolean retval = FALSE;
+
+	/* printf ("search flow look for \"text\" %s\n", info->text); */
+
+	/* first get flow text_len */
+	text_len = 0;
+	while (cur) {
+		if (HTML_OBJECT_TYPE (cur) == HTML_TYPE_TEXTMASTER
+		    || HTML_OBJECT_TYPE (cur) == HTML_TYPE_LINKTEXTMASTER) {
+			text_len += HTML_TEXT (cur)->text_len;
+			end = cur;
+		} else if (HTML_OBJECT_TYPE (cur) != HTML_TYPE_TEXTSLAVE) {
+			break;
+		}
+		cur = (info->forward) ? cur->next : cur->prev;
+	}
+
+	if (text_len > 0) {
+		par = g_new (gchar, text_len+1);
+		par [text_len] = 0;
+
+		pp = (info->forward) ? par : par+text_len;
+
+		/* now fill par with text */
+		head = cur = (info->forward) ? *beg : end;
+		cur = *beg;
+		while (cur) {
+			if (HTML_OBJECT_TYPE (cur) == HTML_TYPE_TEXTMASTER
+			    || HTML_OBJECT_TYPE (cur) == HTML_TYPE_LINKTEXTMASTER) {
+				if (!info->forward) {
+					pp -= HTML_TEXT (cur)->text_len;
+				}
+				strncpy (pp, HTML_TEXT (cur)->text, HTML_TEXT (cur)->text_len);
+				if (info->forward) {
+					pp += HTML_TEXT (cur)->text_len;
+				}
+			} else if (HTML_OBJECT_TYPE (cur) != HTML_TYPE_TEXTSLAVE) {
+				break;
+			}		
+			cur = (info->forward) ? cur->next : cur->prev;
+		}
+
+		printf ("text (%d): %s\n", text_len, par);
+
+		/* set eq_len and pos counters */
+		eq_len = 0;
+		if (info->found) {
+			pos = info->start_pos + ((info->forward) ? 1 : -1);
+		} else {
+			pos = (info->forward) ? 0 : text_len - 1;
+		}
+
+		/* FIXME make shorter text instead */
+		if (!info->forward)
+			par [pos+1] = 0;
+
+		if ((info->forward && pos < text_len)
+		    || (!info->forward && pos>0)) {
+			if (info->reb) {
+				/* regex search */
+				gint rv;
+#ifndef HAVE_GNU_REGEX
+				regmatch_t match;
+				guchar *p=par+pos;
+
+				/* replace &nbsp;'s with spaces */
+				while (*p) {
+					if (*p == ENTITY_NBSP) {
+						*p = ' ';
+					}
+					p += (info->forward) ? 1 : -1;
+				}
+
+				while ((info->forward && pos < text_len)
+				       || (!info->forward && pos >= 0)) {
+					printf ("try match pos: %d %s\n", pos, par + pos);
+					rv = regexec (info->reb,
+						      par + pos,
+						      1, &match, 0);
+					if (rv == 0) {
+						printf ("found! pos: %d start: %d len: %d\n",
+							pos, pos + match.rm_so, match.rm_eo - match.rm_so);
+						search_set_info (head, info, pos + match.rm_so, match.rm_eo - match.rm_so);
+						retval = TRUE;
+						break;
+					}
+					pos += (info->forward) ? 1 : -1;
+				}
+#else
+				rv = re_search (info->reb, par, text_len, pos,
+						(info->forward) ? text_len-pos : -pos, NULL);
+				printf ("search result: %d pos: %d\n", rv, pos);
+				if (rv>=0) {
+					guint found_pos = rv;
+					rv = re_match (info->reb, par, text_len, found_pos, NULL);
+					if (rv < 0) {
+						g_warning ("re_match (...) error");
+					}
+					printf ("found! start: %d len: %d\n", found_pos, rv);
+					search_set_info (head, info, found_pos, rv);
+					retval = TRUE;
+				} else {
+					if (rv < -1) {
+						g_warning ("re_search (...) error");
+					}
+				}
+#endif
+			} else {
+				/* substring search - simple one - could be improved
+				   go thru par and look for info->text */
+				while (par [pos]) {
+					if (info->trans [(guchar) info->text
+							[(info->forward) ? eq_len : info->text_len - eq_len - 1]]
+					    == info->trans [par [pos]]) {
+						eq_len++;
+						if (eq_len == info->text_len) {
+							printf ("found! pos %d\n", pos);
+							search_set_info (head, info, pos - ((info->forward)
+									 ? eq_len-1 : 0), info->text_len);
+							retval=TRUE;
+							break;
+						}
+					} else {
+						pos += (info->forward) ? -eq_len : eq_len;
+						eq_len = 0;
+					}
+					pos += (info->forward) ? 1 : -1;
+				}
+			}
+		}
+		g_free (par);
+	}
+
+	*beg = cur;
+
+	return retval;
+}
+
+static gboolean
+search (HTMLObject *obj, HTMLSearch *info)
+{
+	HTMLClue *clue = HTML_CLUE (obj);
+	HTMLObject *cur;
+	gboolean next = FALSE;
+
+	printf ("search clueflow\n");
+
+	/* does last search end here? */
+	if (info->found) {
+		cur  = HTML_OBJECT (info->found->data);
+		next = TRUE;
+		printf ("search next clueflow continue\n");
+	} else {
+		/* search_next? */
+		if (html_search_child_on_stack (info, obj)) {
+			cur  = html_search_pop (info);
+			cur  = (info->forward) ? cur->next : cur->prev;
+			next = TRUE;
+		} else {
+			/* normal search */
+			cur  = (info->forward) ? clue->head : clue->tail;
+		}
+	}
+	while (cur) {
+		if (HTML_OBJECT_TYPE (cur) == HTML_TYPE_TEXTMASTER
+		    || HTML_OBJECT_TYPE (cur) == HTML_TYPE_LINKTEXTMASTER) {
+			if (search_text (&cur, info)) {
+				return TRUE;
+			}
+		} else {
+			html_search_push (info, cur);
+			if (html_object_search (cur, info)) {
+				return TRUE;
+			}
+			html_search_pop (info);
+			cur = (info->forward) ? cur->next : cur->prev;
+		}
+		if (info->found) {
+			g_list_free (info->found);
+			info->found = NULL;
+			info->start_pos = 0;
+		}
+	}
+
+	if (next) {
+		return html_search_next_parent (info);
+	}
+
+	return FALSE;
+}
+
+static gboolean
+search_next (HTMLObject *obj, HTMLSearch *info)
+{
+	return FALSE;
+}
+
 
 void
 html_clueflow_type_init (void)
@@ -1068,6 +1429,9 @@ html_clueflow_class_init (HTMLClueFlowClass *klass,
 	object_class->save = save;
 	object_class->check_page_split = check_page_split;
 	object_class->check_point = check_point;
+	object_class->append_selection_string = append_selection_string;
+	object_class->search = search;
+	object_class->search_next = search_next;
 
 	klass->get_default_font_style = get_default_font_style;
 
@@ -1159,6 +1523,7 @@ html_clueflow_split (HTMLClueFlow *clue,
 	}
 
 	child->prev = NULL;
+	html_object_change_set (HTML_OBJECT (clue), HTML_CHANGE_ALL);
 
 	/* Put the children into the new clue.  */
 
@@ -1210,30 +1575,52 @@ html_clueflow_set_style (HTMLClueFlow *flow,
 			 HTMLClueFlowStyle style)
 {
 	g_return_if_fail (flow != NULL);
+	g_return_if_fail (engine != NULL);
+	g_return_if_fail (HTML_IS_ENGINE (engine));
 
 	flow->style = style;
 
 	relayout_and_draw (HTML_OBJECT (flow), engine);
 }
 
+HTMLClueFlowStyle
+html_clueflow_get_style (HTMLClueFlow *flow)
+{
+	g_return_val_if_fail (flow != NULL, HTML_CLUEFLOW_STYLE_NORMAL);
+
+	return flow->style;
+}
+
 void
 html_clueflow_set_halignment (HTMLClueFlow *flow,
 			      HTMLEngine *engine,
 			      HTMLHAlignType alignment)
 {
 	g_return_if_fail (flow != NULL);
+	g_return_if_fail (engine != NULL);
+	g_return_if_fail (HTML_IS_ENGINE (engine));
 
 	HTML_CLUE (flow)->halign = alignment;
 
 	relayout_and_draw (HTML_OBJECT (flow), engine);
 }
 
+HTMLHAlignType
+html_clueflow_get_halignment (HTMLClueFlow *flow)
+{
+	g_return_val_if_fail (flow != NULL, HTML_HALIGN_NONE);
+
+	return HTML_CLUE (flow)->halign;
+}
+
 void
 html_clueflow_indent (HTMLClueFlow *flow,
 		      HTMLEngine *engine,
 		      gint indentation)
 {
 	g_return_if_fail (flow != NULL);
+	g_return_if_fail (engine != NULL);
+	g_return_if_fail (HTML_IS_ENGINE (engine));
 
 	if (indentation == 0)
 		return;
@@ -1252,19 +1639,46 @@ html_clueflow_indent (HTMLClueFlow *flow,
 	relayout_for_level_change (flow, engine);
 }
 
+void
+html_clueflow_set_indentation (HTMLClueFlow *flow,
+			       HTMLEngine *engine,
+			       guint8 indentation)
+{
+	g_return_if_fail (flow != NULL);
+	g_return_if_fail (engine != NULL);
+	g_return_if_fail (HTML_IS_ENGINE (engine));
+
+	if (flow->level == indentation)
+		return;
+
+	flow->level = indentation;
+
+	relayout_for_level_change (flow, engine);
+}
+
+guint8
+html_clueflow_get_indentation (HTMLClueFlow *flow)
+{
+	g_return_val_if_fail (flow != NULL, 0);
+
+	return flow->level;
+}
+
 void
 html_clueflow_set_properties (HTMLClueFlow *flow,
 			      HTMLEngine *engine,
 			      HTMLClueFlowStyle style,
-			      guint8 level,
+			      guint8 indentation,
 			      HTMLHAlignType alignment)
 {
 	g_return_if_fail (flow != NULL);
+	g_return_if_fail (engine != NULL);
+	g_return_if_fail (HTML_IS_ENGINE (engine));
 
 	HTML_CLUE (flow)->halign = alignment;
 
 	flow->style = style;
-	flow->level = level;
+	flow->level = indentation;
 
 	relayout_and_draw (HTML_OBJECT (flow), engine);
 }
@@ -1272,15 +1686,15 @@ html_clueflow_set_properties (HTMLClueFlow *flow,
 void
 html_clueflow_get_properties (HTMLClueFlow *flow,
 			      HTMLClueFlowStyle *style_return,
-			      guint8 *level_return,
+			      guint8 *indentation_return,
 			      HTMLHAlignType *alignment_return)
 {
 	g_return_if_fail (flow != NULL);
 
 	if (style_return != NULL)
 		*style_return = flow->style;
-	if (level_return != NULL)
-		*level_return = flow->level;
+	if (indentation_return != NULL)
+		*indentation_return = flow->level;
 	if (alignment_return != NULL)
 		*alignment_return = HTML_CLUE (flow)->halign;
 }
