@@ -49,9 +49,9 @@ debug_location (const HTMLCursor *cursor)
 		return;
 	}
 
-	g_print ("Cursor in %s (%p), offset %d, relative %d\n",
+	g_print ("Cursor in %s (%p), offset %d, position %d\n",
 		 html_type_name (HTML_OBJECT_TYPE (object)),
-		 object, cursor->offset, cursor->relative_position);
+		 object, cursor->offset, cursor->position);
 }
 #else
 #define debug_location(cursor)
@@ -90,7 +90,7 @@ html_cursor_new (void)
 	new->target_x = 0;
 	new->have_target_x = FALSE;
 
-	new->relative_position = 0;
+	new->position = 0;
 
 	return new;
 }
@@ -123,7 +123,7 @@ html_cursor_copy (HTMLCursor *dest,
 	dest->offset = src->offset;
 	dest->target_x = src->target_x;
 	dest->have_target_x = src->have_target_x;
-	dest->relative_position = src->relative_position;
+	dest->position = src->position;
 }
 
 HTMLCursor *
@@ -145,18 +145,6 @@ html_cursor_normalize (HTMLCursor *cursor)
 	normalize (&cursor->object, &cursor->offset);
 }
 
-
-void
-html_cursor_set_position (HTMLCursor *cursor,
-			  HTMLObject *object,
-			  guint offset)
-{
-	g_return_if_fail (cursor != NULL);
-
-	cursor->object = object;
-	cursor->offset = offset;
-}
-
 
 /* This is a gross hack as we don't have a `is_a()' function in the object
    system.  */
@@ -194,6 +182,8 @@ void
 html_cursor_home (HTMLCursor *cursor,
 		  HTMLEngine *engine)
 {
+	HTMLObject *obj;
+
 	g_return_if_fail (cursor != NULL);
 	g_return_if_fail (engine != NULL);
 
@@ -203,17 +193,44 @@ html_cursor_home (HTMLCursor *cursor,
 		return;
 	}
 
-	cursor->object = engine->clue;
-	cursor->offset = 0;
-	cursor->relative_position = 0;
+	obj = engine->clue;
+	while (is_clue (obj))
+		obj = HTML_CLUE (obj)->head;
 
-	html_cursor_forward (cursor, engine);
-	html_cursor_reset_relative (cursor);
+	cursor->object = obj;
+	cursor->offset = 0;
+	cursor->position = 0;
 
 	debug_location (cursor);
 }
 
 
+static HTMLObject *
+forward_object (HTMLObject *obj)
+{
+	obj = next (obj);
+
+	/* Traverse the tree in top-bottom, left-right order until an element
+           that accepts the cursor is found.  */
+
+	while (obj != NULL) {
+		if (is_clue (obj)) {
+			obj = HTML_CLUE (obj)->head;
+			continue;
+		}
+
+		if (html_object_accepts_cursor (obj))
+			break;
+
+		obj = next (obj);
+	}
+
+	if (obj == NULL)
+		return NULL;
+
+	return obj;
+}
+
 static gboolean
 forward (HTMLCursor *cursor,
 	 HTMLEngine *engine)
@@ -234,8 +251,6 @@ forward (HTMLCursor *cursor,
 
 	if (html_object_is_text (obj)) {
 		switch (HTML_OBJECT_TYPE (obj)) {
-		case HTML_TYPE_TEXT:
-		case HTML_TYPE_LINKTEXT:
 		case HTML_TYPE_TEXTMASTER:
 		case HTML_TYPE_LINKTEXTMASTER:
 		{
@@ -247,36 +262,25 @@ forward (HTMLCursor *cursor,
 				break;
 
 			if (offset < text->text_len) {
-				offset++;
-				goto end;
+				cursor->offset++;
+				cursor->position++;
+				return TRUE;
 			}
 
 			if (html_object_next_not_slave (obj) == NULL)
 				offset = 0;
 			else
 				offset = 1;
-
 			break;
 		}
 
 		case HTML_TYPE_TEXTSLAVE:
 			/* Do nothing: go to the next element.  */
-			offset = 0;
 			break;
 
 		default:
 			g_assert_not_reached ();
 		}
-		
-		obj = next (obj);
-		while (obj != NULL && html_object_is_text (obj)) {
-			if (HTML_OBJECT_TYPE (obj) != HTML_TYPE_TEXTMASTER
-			    && HTML_OBJECT_TYPE (obj) != HTML_TYPE_LINKTEXTMASTER) {
-				obj = next (obj);
-			} else {
-				goto end;
-			}
-		}
 	} else if (! is_clue (obj)) {
 		/* Objects that are not a clue or text are always skipped, as
                    they are a unique non-splittable element.  But if the
@@ -284,36 +288,23 @@ forward (HTMLCursor *cursor,
                    so that you have the cursor positioned at the end of the
                    clue.  */
 
-		if (obj->next == NULL && offset == 0) {
-			offset++;
-			goto end;
+		if (offset == 0) {
+			cursor->offset = 1;
+			cursor->position++;
+			return TRUE;
 		}
-
-		obj = next (obj);
-	}
-
-	/* No more text.  Traverse the tree in top-bottom, left-right order
-           until an element that accepts the cursor is found.  */
-
-	while (obj != NULL) {
-		if (is_clue (obj)) {
-			obj = HTML_CLUE (obj)->head;
-			continue;
-		}
-
-		if (html_object_accepts_cursor (obj))
-			break;
-
-		obj = next (obj);
 	}
 
+	obj = forward_object (cursor->object);
 	if (obj == NULL)
 		return FALSE;
 
- end:
 	cursor->object = obj;
-	cursor->offset = offset;
-	cursor->relative_position++;
+	cursor->offset = 0;
+	cursor->position++;
+
+	if (cursor->object->prev != NULL)
+		cursor->offset++;
 
 	return TRUE;
 }
@@ -335,7 +326,59 @@ html_cursor_forward (HTMLCursor *cursor,
 	return retval;
 }
 
+HTMLObject *
+html_object_next_for_cursor (HTMLObject *obj)
+{
+	g_return_val_if_fail (obj != NULL, NULL);
+
+	return forward_object (obj);
+}
+
 
+static HTMLObject *
+backward_object (HTMLObject *obj)
+{
+	while (obj != NULL) {
+		while (obj != NULL && obj->prev == NULL)
+			obj = obj->parent;
+
+		if (obj == NULL)
+			break;
+
+		obj = obj->prev;
+		if (is_clue (obj)) {
+			while (is_clue (obj) && HTML_CLUE (obj)->head != NULL) {
+				obj = HTML_CLUE (obj)->head;
+				while (obj->next != NULL)
+					obj = obj->next;
+			}
+		}
+
+		if (html_object_is_text (obj)) {
+			switch (HTML_OBJECT_TYPE (obj)) {
+			case HTML_TYPE_TEXT:
+			case HTML_TYPE_LINKTEXT:
+			case HTML_TYPE_TEXTMASTER:
+			case HTML_TYPE_LINKTEXTMASTER:
+				return obj;
+
+			case HTML_TYPE_TEXTSLAVE:
+				/* Do nothing: go to the previous element, as
+				   this is not a suitable place for the
+				   cursor.  */
+				break;
+
+			default:
+				g_assert_not_reached ();
+			}
+		} else if (html_object_accepts_cursor (obj)) {
+			return obj;
+		}
+	}
+
+	return NULL;
+}
+
 static gboolean
 backward (HTMLCursor *cursor,
 	  HTMLEngine *engine)
@@ -361,8 +404,9 @@ backward (HTMLCursor *cursor,
 		case HTML_TYPE_TEXTMASTER:
 		case HTML_TYPE_LINKTEXTMASTER:
 			if (offset > 1 || (offset == 1 && obj->prev == NULL)) {
-				offset--;
-				goto end;
+				cursor->offset = offset - 1;
+				cursor->position--;
+				return TRUE;
 			}
 			break;
 
@@ -374,56 +418,25 @@ backward (HTMLCursor *cursor,
 		default:
 			g_assert_not_reached ();
 		}
-	}
-
-	while (obj != NULL) {
-		while (obj != NULL && obj->prev == NULL)
-			obj = obj->parent;
-
-		if (obj == NULL)
-			break;
-
-		obj = obj->prev;
-		if (is_clue (obj)) {
-			while (is_clue (obj) && HTML_CLUE (obj)->head != NULL) {
-				obj = HTML_CLUE (obj)->head;
-				while (obj->next != NULL)
-					obj = obj->next;
-			}
-		}
-
-		if (html_object_is_text (obj)) {
-			switch (HTML_OBJECT_TYPE (obj)) {
-			case HTML_TYPE_TEXT:
-			case HTML_TYPE_LINKTEXT:
-			case HTML_TYPE_TEXTMASTER:
-			case HTML_TYPE_LINKTEXTMASTER:
-				offset = HTML_TEXT (obj)->text_len;
-				goto end;
-
-			case HTML_TYPE_TEXTSLAVE:
-				/* Do nothing: go to the previous element, as
-				   this is not a suitable place for the
-				   cursor.  */
-				break;
-
-			default:
-				g_assert_not_reached ();
-			}
-		} else if (html_object_accepts_cursor (obj)) {
-			break;
-		}
-	}
-
- end:
-	if (obj != NULL) {
-		cursor->object = obj;
-		cursor->offset = offset;
-		cursor->relative_position--;
+	} else if (offset > 0 && obj->prev == NULL) {
+		cursor->offset = 0;
+		cursor->position--;
 		return TRUE;
 	}
 
-	return FALSE;
+	obj = backward_object (obj);
+	if (obj == NULL)
+		return FALSE;
+
+	if (html_object_is_text (obj))
+		cursor->offset = HTML_TEXT (obj)->text_len;
+	else
+		cursor->offset = 1;
+
+	cursor->object = obj;
+	cursor->position--;
+
+	return TRUE;
 }
 
 gboolean
@@ -443,6 +456,14 @@ html_cursor_backward (HTMLCursor *cursor,
 	return retval;
 }
 
+HTMLObject *
+html_object_prev_for_cursor (HTMLObject *object)
+{
+	g_return_val_if_fail (object != NULL, NULL);
+
+	return backward_object (object);
+}
+
 
 gboolean
 html_cursor_up (HTMLCursor *cursor,
@@ -517,7 +538,7 @@ html_cursor_up (HTMLCursor *cursor,
 			if (prev_y == y && target_x - x >= prev_x - target_x) {
 				cursor->object = prev_cursor.object;
 				cursor->offset = prev_cursor.offset;
-				cursor->relative_position = prev_cursor.relative_position;
+				cursor->position = prev_cursor.position;
 			}
 
 			debug_location (cursor);
@@ -530,7 +551,6 @@ html_cursor_up (HTMLCursor *cursor,
 gboolean
 html_cursor_down (HTMLCursor *cursor,
 		  HTMLEngine *engine)
-
 {
 	HTMLCursor orig_cursor;
 	HTMLCursor prev_cursor;
@@ -598,7 +618,7 @@ html_cursor_down (HTMLCursor *cursor,
 			if (prev_y == y && x - target_x >= target_x - prev_x) {
 				cursor->object = prev_cursor.object;
 				cursor->offset = prev_cursor.offset;
-				cursor->relative_position = prev_cursor.relative_position;
+				cursor->position = prev_cursor.position;
 			}
 
 			debug_location (cursor);
@@ -654,98 +674,6 @@ html_cursor_jump_to (HTMLCursor *cursor,
 	return FALSE;
 }
 
-
-/* Comparison.  */
-gboolean
-html_cursor_equal (HTMLCursor *a, HTMLCursor *b)
-{
-	g_return_val_if_fail (a != NULL, FALSE);
-	g_return_val_if_fail (b != NULL, FALSE);
-
-	return a->object == b->object && a->offset == b->offset;
-}
-
-
-/* Relative coordinate support.  This is used to keep track of how far
-   we have moved with the pointer, e.g. for undo/redo support.  */
-
-gint
-html_cursor_get_relative (HTMLCursor *cursor)
-{
-	g_return_val_if_fail (cursor != NULL, 0);
-
-	return cursor->relative_position;
-}
-
-void
-html_cursor_set_relative (HTMLCursor *cursor,
-			  gint relative_position)
-{
-	g_return_if_fail (cursor != NULL);
-
-	cursor->relative_position = relative_position;
-}
-
-void
-html_cursor_reset_relative (HTMLCursor *cursor)
-{
-	g_return_if_fail (cursor != NULL);
-
-	cursor->relative_position = 0;
-}
-
-void
-html_cursor_goto_zero (HTMLCursor *cursor,
-		       HTMLEngine *engine)
-{
-	g_return_if_fail (cursor != NULL);
-	g_return_if_fail (engine != NULL);
-	g_return_if_fail (HTML_IS_ENGINE (engine));
-
-	if (cursor->relative_position == 0)
-		return;
-
-	if (cursor->relative_position > 0) {
-		while (cursor->relative_position > 0)
-			backward (cursor, engine);
-	} else {
-		while (cursor->relative_position < 0)
-			forward (cursor, engine);
-	}
-
-	debug_location (cursor);
-}
-
-
-gchar
-html_cursor_get_current_char (HTMLCursor *cursor)
-{
-	HTMLObject *next;
-
-	g_return_val_if_fail (cursor != NULL, 0);
-
-	if (! html_object_is_text (cursor->object)) {
-		if (cursor->offset == 0)
-			return 0;
-
-		next = cursor->object->next;
-		if (next != NULL && html_object_is_text (next))
-			return HTML_TEXT (next)->text[0];
-
-		return 0;
-	}
-
-	if (cursor->offset < HTML_TEXT (cursor->object)->text_len)
-		return HTML_TEXT (cursor->object)->text[cursor->offset];
-
-	next = html_object_next_not_slave (cursor->object);
-
-	if (next == NULL || ! html_object_is_text (next))
-		return 0;
-
-	return HTML_TEXT (next)->text[0];
-}
-
 
 /* Complex cursor movement commands.  */
 
@@ -838,3 +766,96 @@ html_cursor_beginning_of_line (HTMLCursor *cursor,
 		html_cursor_copy (&prev_cursor, cursor);
 	}
 }
+
+
+gint
+html_cursor_get_position (HTMLCursor *cursor)
+{
+	g_return_val_if_fail (cursor != NULL, 0);
+
+	return cursor->position;
+}
+
+void
+html_cursor_jump_to_position (HTMLCursor *cursor,
+			      HTMLEngine *engine,
+			      gint position)
+{
+	g_return_if_fail (cursor != NULL);
+	g_return_if_fail (position >= 0);
+
+	if (cursor->position < position) {
+		while (cursor->position < position) {
+			if (! forward (cursor, engine))
+				break;
+		}
+	} else if (cursor->position > position) {
+		while (cursor->position > position) {
+			if (! backward (cursor, engine))
+				break;
+		}
+	}
+}
+
+
+/* Comparison.  */
+
+gboolean
+html_cursor_equal (const HTMLCursor *a,
+		   const HTMLCursor *b)
+{
+	g_return_val_if_fail (a != NULL, FALSE);
+	g_return_val_if_fail (b != NULL, FALSE);
+
+	return a->object == b->object && a->offset == b->offset;
+}
+
+gboolean
+html_cursor_precedes (const HTMLCursor *a,
+		      const HTMLCursor *b)
+{
+	g_return_val_if_fail (a != NULL, FALSE);
+	g_return_val_if_fail (b != NULL, FALSE);
+
+	return a->position < b->position;
+}
+
+gboolean
+html_cursor_follows (const HTMLCursor *a,
+		     const HTMLCursor *b)
+{
+	g_return_val_if_fail (a != NULL, FALSE);
+	g_return_val_if_fail (b != NULL, FALSE);
+
+	return a->position > b->position;
+}
+
+
+gchar
+html_cursor_get_current_char (const HTMLCursor *cursor)
+{
+	HTMLObject *next;
+
+	g_return_val_if_fail (cursor != NULL, 0);
+
+	if (! html_object_is_text (cursor->object)) {
+		if (cursor->offset == 0)
+			return 0;
+
+		next = cursor->object->next;
+		if (next != NULL && html_object_is_text (next))
+			return HTML_TEXT (next)->text[0];
+
+		return 0;
+	}
+
+	if (cursor->offset < HTML_TEXT (cursor->object)->text_len)
+		return HTML_TEXT (cursor->object)->text[cursor->offset];
+
+	next = html_object_next_not_slave (cursor->object);
+
+	if (next == NULL || ! html_object_is_text (next))
+		return 0;
+
+	return HTML_TEXT (next)->text[0];
+}
