@@ -77,6 +77,7 @@ draw (HTMLObject *o,
 		}
 		element->abs_x = new_x;
 		element->abs_y = new_y;
+		gtk_widget_queue_draw (element->widget);
 	}
 }
 
@@ -91,17 +92,101 @@ destroy (HTMLObject *o)
 		g_free(element->name);
 	if(element->value)
 		g_free(element->value);
-	if(element->widget)
-		gtk_widget_destroy(element->widget);
+	if(element->widget) {
+		if (element->widget->parent)
+			gtk_widget_destroy (element->widget);
+		else
+			gtk_widget_unref (element->widget);
+		}
 
-	HTML_OBJECT_CLASS (&html_object_class)->destroy (o);
+	HTML_OBJECT_CLASS (parent_class)->destroy (o);
 }
 
 static void
 reset (HTMLEmbedded *e)
 {
+	/* Nothing to do?  */
 }
 
+static gint
+calc_min_width (HTMLObject *self,
+		HTMLPainter *painter)
+{
+	HTMLEmbedded *emb = HTML_EMBEDDED (self);
+	GtkWidget *widget;
+	gint pixel_size;
+	gint min_width;
+
+	widget = HTML_EMBEDDED (self)->widget;
+	if (widget == NULL)
+		return 0;
+
+	pixel_size = html_painter_get_pixel_size (painter);
+
+	if (HTML_EMBEDDED (self)->allocated) {
+		/* set width and height to minimize updates */
+		emb->width  = widget->allocation.width;
+		emb->height = widget->allocation.height;
+		min_width = widget->allocation.width * pixel_size;
+	} else {
+		GtkRequisition req;
+
+		gtk_widget_size_request (widget, &req);
+
+		/* set width and height to minimize updates */
+		emb->width   = req.width;
+		emb->height  = req.height;
+		min_width = req.width * pixel_size;
+	}
+
+	return min_width;
+}
+
+static gboolean
+calc_size (HTMLObject *self,
+	   HTMLPainter *painter)
+{
+	GtkWidget *widget;
+	HTMLEmbedded *emb = HTML_EMBEDDED (self);
+	gint pixel_size;
+	gint old_width, old_ascent;
+
+	widget = emb->widget;
+	if (widget == NULL)
+		return FALSE;
+
+	pixel_size = html_painter_get_pixel_size (painter);
+
+	old_width = self->width;
+	old_ascent = self->ascent;
+
+	if (HTML_EMBEDDED (self)->allocated) {
+		/* set width and height to minimize updates */
+		emb->width   = widget->allocation.width;
+		emb->height  = widget->allocation.height;
+		self->width  = widget->allocation.width * pixel_size;
+		self->ascent = widget->allocation.height * pixel_size;
+	} else {
+		GtkRequisition req;
+
+		gtk_widget_size_request (widget, &req);
+		/* set width and height to minimize updates */
+		emb->width   = req.width;
+		emb->height  = req.height;
+		self->width  = req.width * pixel_size;
+		self->ascent = req.height * pixel_size;
+	}
+
+	/* This never changes.  */
+	self->descent = 0;
+
+	if (old_width != self->width || old_ascent != self->ascent)
+		return TRUE;
+
+	return FALSE;
+}
+
+
 void
 html_embedded_reset (HTMLEmbedded *e)
 {
@@ -197,6 +282,8 @@ html_embedded_class_init (HTMLEmbeddedClass *klass,
 	object_class->destroy = destroy;
 	object_class->copy = copy;
 	object_class->draw = draw;
+	object_class->calc_size = calc_size;
+	object_class->calc_min_width = calc_min_width;
 
 	parent_class = &html_object_class;
 }
@@ -224,6 +311,9 @@ html_embedded_init (HTMLEmbedded *element,
 		element->value = g_strdup("");
 	element->widget = NULL;
 	element->parent = parent;
+	element->width  = 0;
+	element->height = 0;
+	element->allocated = FALSE;
 
 	element->abs_x = element->abs_y = -1;
 }
@@ -246,6 +336,7 @@ html_embedded_size_recalc(HTMLEmbedded *em)
 	} else {
 		o->descent = 0;
 	}
+
 	o->ascent = req.height - o->descent;
 }
 
@@ -272,3 +363,37 @@ html_embedded_new_widget(GtkWidget *parent, GtkHTMLEmbedded *eb)
 	return em;
 }
 
+static void
+allocate (GtkWidget *w, GtkAllocation  *allocation, HTMLEmbedded *e)
+{
+	if (e->width != allocation->width || e->height != allocation->height) {
+		if (e->width != allocation->width) {
+			html_object_change_set (HTML_OBJECT (e), HTML_CHANGE_MIN_WIDTH);
+			e->width = allocation->width;
+		}
+		e->height = allocation->height;
+
+		g_assert (GTK_IS_HTML (w->parent));
+		html_engine_schedule_update (GTK_HTML (w->parent)->engine);
+	}
+
+	e->allocated = TRUE;
+}
+
+static void
+request (GtkWidget *w, GtkRequisition *req, HTMLEmbedded *e)
+{
+	e->allocated = FALSE;
+}
+
+void
+html_embedded_set_widget (HTMLEmbedded *e, GtkWidget *w)
+{
+	e->widget = w;
+
+	gtk_signal_connect (GTK_OBJECT (w), "size_allocate",
+			    GTK_SIGNAL_FUNC (allocate), e);
+	gtk_signal_connect (GTK_OBJECT (w), "size_request",
+			    GTK_SIGNAL_FUNC (request), e);
+
+}
