@@ -29,8 +29,10 @@
 #include "htmlcursor.h"
 #include "htmltext.h"
 #include "htmltextslave.h"
+#include "htmlimage.h"
 
 #include "htmlengine-edit.h"
+#include "htmlengine-edit-paste.h"
 
 
 void
@@ -41,8 +43,9 @@ html_engine_undo (HTMLEngine *e)
 	g_return_if_fail (e != NULL);
 	g_return_if_fail (HTML_IS_ENGINE (e));
 	g_return_if_fail (e->undo != NULL);
+	g_return_if_fail (e->editable);
 
-	g_warning ("*Undo!*");
+	html_engine_unselect_all (e, TRUE);
 
 	html_engine_freeze (e);
 
@@ -61,7 +64,7 @@ html_engine_redo (HTMLEngine *e)
 	g_return_if_fail (HTML_IS_ENGINE (e));
 	g_return_if_fail (e->undo != NULL);
 
-	g_warning ("*Redo!*");
+	html_engine_unselect_all (e, TRUE);
 
 	html_engine_freeze (e);
 
@@ -70,3 +73,21 @@ html_engine_redo (HTMLEngine *e)
 
 	html_engine_thaw (e);
 }
+
+
+void
+html_engine_set_mark (HTMLEngine *e)
+{
+	g_return_if_fail (e != NULL);
+	g_return_if_fail (HTML_IS_ENGINE (e));
+	g_return_if_fail (e->editable);
+
+	if (e->mark != NULL)
+		html_engine_unselect_all (e, TRUE);
+
+	e->mark = html_cursor_dup (e->cursor);
+	e->active_selection = TRUE;
+
+	html_engine_edit_selection_updater_reset (e->selection_updater);
+	html_engine_edit_selection_updater_schedule (e->selection_updater);
+}
