@@ -319,37 +319,51 @@ static void
 draw_panel (HTMLPainter *painter,
 	    gint x, gint y,
 	    gint width, gint height,
-	    gint inset,
+	    GtkHTMLEtchStyle inset,
 	    gint bordersize)
 {
 	HTMLGdkPainter *gdk_painter;
-	GdkColor *col1, *col2;
+	GdkColor *col1 = NULL, *col2 = NULL;
 
 	gdk_painter = HTML_GDK_PAINTER (painter);
 
-	if (inset) {
-		col1 = &gdk_painter->dark;
-		col2 = &gdk_painter->light;
-	}
-	else {
+	switch (inset) {
+	case GTK_HTML_ETCH_NONE:
+		/* use the current pen color */
+		col1 = NULL;
+		col2 = NULL;
+		break;
+	case GTK_HTML_ETCH_OUT:
 		col1 = &gdk_painter->light;
 		col2 = &gdk_painter->dark;
+		break;
+	default:
+	case GTK_HTML_ETCH_IN:
+		col1 = &gdk_painter->dark;
+		col2 = &gdk_painter->light;
+		break;
 	}
-
+	
 	x -= gdk_painter->x1;
 	y -= gdk_painter->y1;
 	
 	while (bordersize > 0) {
-		gdk_gc_set_foreground (gdk_painter->gc, col1);
-		gdk_draw_line (gdk_painter->pixmap, gdk_painter->gc,
-			       x, y, x + width - 2, y);
-		gdk_draw_line (gdk_painter->pixmap, gdk_painter->gc,
-			       x, y, x, y + height - 1);
-		gdk_gc_set_foreground (gdk_painter->gc, col2);
+		if (col2) {
+			gdk_gc_set_foreground (gdk_painter->gc, col2);
+		}
+
 		gdk_draw_line (gdk_painter->pixmap, gdk_painter->gc,
 			       x + width - 1, y, x + width - 1, y + height - 1);
 		gdk_draw_line (gdk_painter->pixmap, gdk_painter->gc,
 			       x + 1, y + height - 1, x + width - 1, y + height - 1);
+		if (col1) {
+			gdk_gc_set_foreground (gdk_painter->gc, col1);
+		}
+
+		gdk_draw_line (gdk_painter->pixmap, gdk_painter->gc,
+			       x, y, x + width - 2, y);
+		gdk_draw_line (gdk_painter->pixmap, gdk_painter->gc,
+			       x, y, x, y + height - 1);
 		bordersize--;
 		x++;
 		y++;
@@ -359,24 +373,115 @@ draw_panel (HTMLPainter *painter,
 }
 
 static void
-draw_background_pixmap (HTMLPainter *painter,
-			gint x, gint y, 
-			GdkPixbuf *pixbuf,
-			gint pix_width, gint pix_height)
+draw_background (HTMLPainter *painter,
+		 const GdkColor *color,
+		 GdkPixbuf *pixbuf,
+		 gint x, gint y, 
+		 gint width, gint height,
+		 gint tile_x, gint tile_y)
 {
 	HTMLGdkPainter *gdk_painter;
+	gint pw;
+	gint ph;
+	gint tile_width, tile_height;
+	gint w, h;
+
+	tile_width = width + tile_x;
+	tile_height = height + tile_y;
+
+	if (!color && !pixbuf)
+		return;
 
 	gdk_painter = HTML_GDK_PAINTER (painter);
 
-	gdk_pixbuf_render_to_drawable_alpha (pixbuf, gdk_painter->pixmap,
-					     0, 0,
-					     x - gdk_painter->x1, y - gdk_painter->y1, 
-					     pix_width ? pix_width : pixbuf->art_pixbuf->width,
-					     pix_height ? pix_height : pixbuf->art_pixbuf->height,
-					     GDK_PIXBUF_ALPHA_BILEVEL,
-					     128,
-					     GDK_RGB_DITHER_NORMAL,
-					     x, y);
+	if (color) {
+		gdk_gc_set_foreground (gdk_painter->gc, color);
+		gdk_draw_rectangle (gdk_painter->pixmap, gdk_painter->gc,
+				    TRUE, x - gdk_painter->x1, y - gdk_painter->y1,
+				    width, height);	
+		
+	}
+
+	if (!pixbuf)
+		return;
+
+	pw = gdk_pixbuf_get_width (pixbuf);
+	ph = gdk_pixbuf_get_height (pixbuf);
+
+	/* do tiling */
+	if (tile_width > pw || tile_height > ph) {
+		GdkPixmap * pixmap;
+		gint cw, ch, cx, cy;
+		GdkGC *gc;
+		GdkBitmap *bitmap = NULL;
+
+		gc = gdk_gc_new (gdk_painter->window);
+		
+		if (gdk_pixbuf_get_has_alpha (pixbuf)) {
+			
+			/* Right now we only support GDK_PIXBUF_ALPHA_BILEVEL, so we
+			 * unconditionally create the clipping mask.
+			 */
+			
+			bitmap = gdk_pixmap_new (NULL, pw, ph, 1);
+			gdk_pixbuf_render_threshold_alpha (pixbuf, bitmap,
+							   0, 0,
+							   0, 0,
+							   pw, ph,
+							   128);
+			
+			gdk_gc_set_clip_mask (gc, bitmap);
+		}
+			
+		pixmap = gdk_pixmap_new (gdk_painter->window, pw, ph, -1);
+		gdk_pixbuf_render_to_drawable (pixbuf, pixmap, gc,
+					       0, 0,
+					       0, 0, 
+					       pw, ph,
+					       GDK_RGB_DITHER_NORMAL,
+					       x, y);
+
+		cy = y;
+		ch = height;
+		h = tile_y % ph;
+		while (ch > 0) {
+			cx = x;
+			cw = width;
+			w = tile_x % pw;
+			while (cw > 0) {
+
+				if (bitmap)
+					gdk_gc_set_clip_origin (gc, cx - w - gdk_painter->x1, cy - h - gdk_painter->y1);
+				
+				
+				gdk_draw_pixmap (gdk_painter->pixmap, gc, pixmap,
+						 w, h, cx - gdk_painter->x1, cy - gdk_painter->y1,
+						 (cw >= pw -w) ? pw -w: cw,
+						 (ch >= ph -h) ? ph -h: ch);
+				cw -= pw - w;
+				cx += pw - w;
+				w = 0;
+			}
+			ch -= ph - h;
+			cy += ph - h;
+			h = 0;
+		}
+		gdk_pixmap_unref (pixmap);
+
+		if (bitmap)
+			gdk_bitmap_unref (bitmap);
+
+		gdk_gc_unref (gc);
+	} else {
+		gdk_pixbuf_render_to_drawable_alpha (pixbuf, gdk_painter->pixmap,
+						     tile_x, tile_y,
+						     x - gdk_painter->x1, y - gdk_painter->y1, 
+						     width, height,
+						     GDK_PIXBUF_ALPHA_BILEVEL,
+						     128,
+						     GDK_RGB_DITHER_NORMAL,
+						     x, y);
+	}
 }
 
 static GdkPixbuf *
@@ -384,19 +489,13 @@ create_temporary_pixbuf (GdkPixbuf *src,
 			 gint clip_width, gint clip_height)
 {
 	GdkPixbuf *pixbuf;
-	gint src_width, src_height;
 	gboolean has_alpha;
-	guint n_channels;
 	guint bits_per_sample;
 
-	src_width = gdk_pixbuf_get_width (src);
-	src_height = gdk_pixbuf_get_height (src);
-
 	has_alpha = gdk_pixbuf_get_has_alpha (src);
-	n_channels = gdk_pixbuf_get_n_channels (src);
 	bits_per_sample = gdk_pixbuf_get_bits_per_sample (src);
 
-	pixbuf = gdk_pixbuf_new (ART_PIX_RGB, has_alpha, bits_per_sample, clip_width, clip_height);
+	pixbuf = gdk_pixbuf_new (GDK_COLORSPACE_RGB, has_alpha, bits_per_sample, clip_width, clip_height);
 
 	return pixbuf;
 }
@@ -444,7 +543,8 @@ draw_pixmap (HTMLPainter *painter,
 	paint_width = paint.x1 - paint.x0;
 	paint_height = paint.y1 - paint.y0;
 
-	if (scale_width == orig_width && scale_height == orig_height && color == NULL) {
+
+	if (scale_width == orig_width && scale_height == orig_height && color == NULL && (!gdk_painter->alpha)) {
 		gdk_pixbuf_render_to_drawable_alpha (pixbuf, gdk_painter->pixmap,
 						     paint.x0 - image.x0,
 						     paint.y0 - image.y0,
@@ -459,21 +559,31 @@ draw_pixmap (HTMLPainter *painter,
 		return;
 	}
 
-	tmp_pixbuf = create_temporary_pixbuf (pixbuf,
-					      paint_width,
-					      paint_height);
+	if (gdk_pixbuf_get_has_alpha (pixbuf) && gdk_painter->alpha) {
+	    tmp_pixbuf = gdk_pixbuf_get_from_drawable(NULL,
+						      gdk_painter->pixmap,
+						      gdk_window_get_colormap (gdk_painter->window),
+						      paint.x0 - clip.x0, 
+						      paint.y0 - clip.y0,
+						      0, 0, paint_width, paint_height);
+	} else {
+		tmp_pixbuf = create_temporary_pixbuf (pixbuf,
+						      paint_width,
+						      paint_height);
+	}
+
 	if (tmp_pixbuf == NULL)
 		return;
 
-	gdk_pixbuf_scale (pixbuf, tmp_pixbuf,
-			  0, 0,
-			  paint_width,
-			  paint_height,
-			  -(paint.x0 - image.x0), 
-			  -(paint.y0 - image.y0),
-			  (double) scale_width/ (double) orig_width,
-			  (double) scale_height/ (double) orig_height,
-			  ART_FILTER_BILINEAR);
+	gdk_pixbuf_composite (pixbuf, tmp_pixbuf,
+			      0,
+			      0,
+			      paint_width, paint_height,
+			      (double)-(paint.x0 - image.x0), 
+			      (double)-(paint.y0 - image.y0),
+			      (gdouble) scale_width/ (gdouble) orig_width,
+			      (gdouble) scale_height/ (gdouble) orig_height,
+			      GDK_INTERP_BILINEAR, 255);
 
 	if (color != NULL) {
 		guchar *p, *q;
@@ -693,6 +803,7 @@ init (GtkObject *object)
 
 	gdk_painter->window = NULL;
 
+	gdk_painter->alpha = TRUE;
 	gdk_painter->gc = NULL;
 
 	gdk_painter->double_buffer = TRUE;
@@ -741,7 +852,7 @@ class_init (GtkObjectClass *object_class)
 	painter_class->draw_shade_line = draw_shade_line;
 	painter_class->draw_panel = draw_panel;
 	painter_class->set_clip_rectangle = set_clip_rectangle;
-	painter_class->draw_background_pixmap = draw_background_pixmap;
+	painter_class->draw_background = draw_background;
 	painter_class->set_color_set = set_color_set;
 	painter_class->get_pixel_size = get_pixel_size;
 
