@@ -33,6 +33,7 @@ typedef struct _HTMLImage HTMLImage;
 typedef struct _HTMLImageClass HTMLImageClass;
 typedef struct _HTMLImagePointer HTMLImagePointer;
 typedef struct _HTMLImageFactory HTMLImageFactory;
+typedef struct _HTMLImageAnimation HTMLImageAnimation;
 
 struct _HTMLImagePointer {
 	gchar *url;
@@ -43,16 +44,49 @@ struct _HTMLImagePointer {
 	HTMLImageFactory *factory;
 };
 
+struct _HTMLImageAnimation {
+	/* stuff for animated image */
+	/* draw coordination */
+	gint x;
+	gint y;
+
+	/* remember engine offset */
+	gint ex, ey;
+
+	/* current frame */
+	GList *cur_frame;
+	gint cur_n;
+
+	/* animation timeout function */
+	gint timeout;
+
+	/* helper buffer */
+	GdkPixbuf *pixbuf;
+
+	/* active draw flag */
+	gint active;
+};
+
 #define HTML_IMAGE(x) ((HTMLImage *)(x))
 
 struct _HTMLImage {
 	HTMLObject object;
         HTMLImagePointer *image_ptr;
+	HTMLImageAnimation *animation;
 
-	gint border;
+	GdkColor color;
+	gboolean color_allocated: 1;
+	gboolean have_color: 1;
 
-	gint specified_width;
-	gint specified_height;
+	gint8 border;
+
+	gint16 specified_width;
+	gint16 specified_height;
+
+	gint8 hspace;
+	gint8 vspace;
+
+	gint8 valign;		/* HTMLVAlignType */
 
 	gchar *url;
 	gchar *target;
@@ -73,26 +107,38 @@ void        html_image_class_init  (HTMLImageClass   *klass,
 void        html_image_init        (HTMLImage        *image,
 				    HTMLImageClass   *klass,
 				    HTMLImageFactory *imf,
-				    gchar            *filename,
+				    const gchar      *filename,
 				    const gchar      *url,
 				    const gchar      *target,
-				    gint              width,
-				    gint              height,
-				    gint              percent,
-				    gint              border);
+				    gint16            width,
+				    gint16            height,
+				    gint8             percent,
+				    gint8             border,
+				    const GdkColor   *border_color,
+				    HTMLVAlignType    valign);
 HTMLObject *html_image_new         (HTMLImageFactory *imf,
-				    gchar            *filename,
+				    const gchar      *filename,
 				    const gchar      *url,
 				    const gchar      *target,
-				    gint              width,
-				    gint              height,
-				    gint              percent,
-				    gint              border);
+				    gint16            width,
+				    gint16            height,
+				    gint8             percent,
+				    gint8             border,
+				    const GdkColor   *border_color,
+				    HTMLVAlignType    valign);
+
+void         html_image_set_size           (HTMLImage *image, gint w, gint percent, gint h);
+void         html_image_set_spacing        (HTMLImage *image, gint hspace, gint vspace);
+void         html_image_set_filename       (HTMLImage *image, const gchar *file);
+void         html_image_set_valign         (HTMLImage *image, HTMLVAlignType valign);
+void         html_image_set_border         (HTMLImage *image, gint border);
 
 /* FIXME move to htmlimagefactory.c */
 HTMLImageFactory *html_image_factory_new      (HTMLEngine       *e);
 void              html_image_factory_free     (HTMLImageFactory *factory);
 void              html_image_factory_cleanup  (HTMLImageFactory *factory); /* Does gc etc. - removes unused image entries */
+void              html_image_factory_stop_animations (HTMLImageFactory *factory);
+void              html_image_factory_deactivate_animations (HTMLImageFactory *factory);
 
 HTMLImagePointer *html_image_factory_register    (HTMLImageFactory *factory,
 						  HTMLImage        *i,
