@@ -66,12 +66,15 @@ copy (HTMLObject *self,
 	dest->ascent = self->ascent;
 	dest->descent = self->descent;
 	dest->width = self->width;
+	dest->min_width = self->min_width;
 	dest->max_width = self->max_width;
+	dest->nb_width = self->nb_width;
 	dest->percent = self->percent;
 	dest->flags = self->flags;
 	dest->redraw_pending = self->redraw_pending;
 	dest->selected = self->selected;
 	dest->free_pending = FALSE;
+	dest->change = self->change;
 }
 
 static void
@@ -175,6 +178,9 @@ get_right_margin (HTMLObject *self, gint y)
 static void
 reset (HTMLObject *o)
 {
+	o->width = 0;
+	o->ascent = 0;
+	o->descent = 0;
 }
 
 static const gchar *
@@ -240,6 +246,8 @@ relayout (HTMLObject *self,
 	prev_ascent = self->ascent;
 	prev_descent = self->descent;
 
+	/* Notice that this will reset ascent and descent which we
+           need afterwards.  Yeah, yuck, bleargh.  */
 	html_object_reset (self);
 
 	/* Crappy hack to make crappy htmlclueflow.c happy.  */
@@ -248,7 +256,7 @@ relayout (HTMLObject *self,
 			   "Y value < height of object!\n");
 		self->y = 0;
 	} else {
-		self->y -= self->ascent + self->descent;
+		self->y -= prev_ascent + prev_descent;
 	}
 
 	changed = html_object_calc_size (self, engine->painter);
@@ -305,6 +313,12 @@ relayout (HTMLObject *self,
 	return TRUE;
 }
 
+static HTMLVAlignType
+get_valign (HTMLObject *self)
+{
+	return HTML_VALIGN_BOTTOM;
+}
+
 static gboolean
 accepts_cursor (HTMLObject *self)
 {
@@ -369,6 +383,8 @@ static HTMLObject *
 get_selection (HTMLObject *self,
 	       guint *size_return)
 {
+	HTMLObject *copy;
+
 	if (! self->selected) {
 		if (size_return != NULL)
 			*size_return = 0;
@@ -378,7 +394,16 @@ get_selection (HTMLObject *self,
 	if (size_return != NULL)
 		*size_return = 1;
 
-	return html_object_dup (self);
+	copy = html_object_dup (self);
+	html_object_select_range (copy, NULL, 0, 0, FALSE);
+
+	return copy;
+}
+
+static void
+append_selection_string (HTMLObject *self,
+			 GString *buffer)
+{
 }
 
 static void
@@ -409,6 +434,13 @@ check_page_split (HTMLObject *self,
 	return y;
 }
 
+static gboolean
+search (HTMLObject *self, HTMLSearch *info)
+{
+	/* not found by default */
+	return FALSE;
+}
+
 
 /* Class initialization.  */
 
@@ -451,15 +483,19 @@ html_object_class_init (HTMLObjectClass *klass,
 	klass->set_bg_color = set_bg_color;
 	klass->check_point = check_point;
 	klass->relayout = relayout;
+	klass->get_valign = get_valign;
 	klass->accepts_cursor = accepts_cursor;
 	klass->get_cursor = get_cursor;
 	klass->get_cursor_base = get_cursor_base;
 	klass->select_range = select_range;
 	klass->get_selection = get_selection;
+	klass->append_selection_string = append_selection_string;
 	klass->forall = forall;
 	klass->is_container = is_container;
 	klass->save = save;
 	klass->check_page_split = check_page_split;
+	klass->search = search;
+	klass->search_next = search;
 }
 
 void
@@ -472,6 +508,9 @@ html_object_init (HTMLObject *o,
 	o->prev = NULL;
 	o->next = NULL;
 
+	/* we don't have any info cached in the beginning */
+	o->change = HTML_CHANGE_ALL;
+
 	o->x = 0;
 	o->y = 0;
 
@@ -542,6 +581,24 @@ html_object_calc_abs_position (HTMLObject *o,
 	}
 }
 
+void
+html_object_calc_intersection (HTMLObject *o, ArtIRect *intersection, gint x, gint y, gint width, gint height)
+{
+	ArtIRect clip, area;
+	
+	area.x0 = o->x;
+	area.x1 = o->x + o->width;
+	area.y0 = o->y - o->ascent;
+	area.y1 = o->y + o->descent;
+
+	clip.x0 = x;
+	clip.x1 = x + width;
+	clip.y0 = y;
+	clip.y1 = y + height;
+
+	art_irect_intersect (intersection, &clip, &area);
+}
+
 
 /* Virtual methods.  */
 
@@ -644,7 +701,11 @@ gint
 html_object_calc_min_width (HTMLObject *o,
 			    HTMLPainter *painter)
 {
-	return (* HO_CLASS (o)->calc_min_width) (o, painter);
+	if (o->change & HTML_CHANGE_MIN_WIDTH) {
+		o->min_width = (* HO_CLASS (o)->calc_min_width) (o, painter);
+		o->change &= ~HTML_CHANGE_MIN_WIDTH;
+	}
+	return o->min_width;
 }
 
 gint
@@ -698,6 +759,14 @@ html_object_relayout (HTMLObject *self,
 	return (* HO_CLASS (self)->relayout) (self, engine, child);
 }
 
+HTMLVAlignType
+html_object_get_valign (HTMLObject *self)
+{
+	g_return_val_if_fail (self != NULL, HTML_VALIGN_BOTTOM);
+
+	return (* HO_CLASS (self)->get_valign) (self);
+}
+
 gboolean
 html_object_accepts_cursor (HTMLObject *self)
 {
@@ -745,6 +814,16 @@ html_object_get_selection (HTMLObject *self,
 	return (* HO_CLASS (self)->get_selection) (self, size_return);
 }
 
+void
+html_object_append_selection_string (HTMLObject *self,
+				     GString *buffer)
+{
+	g_return_if_fail (self != NULL);
+	g_return_if_fail (buffer != NULL);
+
+	(* HO_CLASS (self)->append_selection_string) (self, buffer);
+}
+
 
 void
 html_object_forall (HTMLObject *self,
@@ -825,3 +904,29 @@ html_object_check_page_split  (HTMLObject *self,
 	return (* HO_CLASS (self)->check_page_split) (self, y);
 }
 
+void
+html_object_change_set (HTMLObject *self, HTMLChangeFlags f)
+{
+	HTMLObject *obj = self;
+
+	g_assert (self != NULL);
+
+	if (f != HTML_CHANGE_NONE) {
+		while (obj) {
+			obj->change |= f;
+			obj = obj->parent;
+		}
+	}
+}
+
+gboolean
+html_object_search (HTMLObject *self, HTMLSearch *info)
+{
+	return (* HO_CLASS (self)->search) (self, info);
+}
+
+gboolean
+html_object_search_next (HTMLObject *self, HTMLSearch *info)
+{
+	return (* HO_CLASS (self)->search_next) (self, info);
+}
