@@ -27,6 +27,7 @@
 #include "htmltext.h"
 #include "htmlclueflow.h"
 #include "htmlcursor.h"
+#include "htmlentity.h"
 
 
 HTMLTextClass html_text_class;
@@ -35,89 +36,6 @@ static HTMLObjectClass *parent_class = NULL;
 #define HT_CLASS(x) HTML_TEXT_CLASS (HTML_OBJECT (x)->klass)
 
 
-/* Calculcate the length of the string after collapsing multiple spaces into
-   single ones and, optionally, removing leading/trailing space completely.  */
-static guint
-calc_actual_len (const gchar *s,
-		 guint len,
-		 gboolean remove_leading_space,
-		 gboolean remove_trailing_space,
-		 const gchar **first_return)
-{
-	guint i;
-	guint actual_len;
-
-	if (len == 0)
-		return 0;
-
-	i = 0;
-	actual_len = 0;
-
-	if (remove_leading_space) {
-		while (i < len && s[i] == ' ')
-			i++;
-		len -= i;
-		if (len == 0)
-			return 0;
-	} 
-
-	*first_return = s + i;
-
-	if (remove_trailing_space) {
-		while (len > 0 && s[len - 1] == ' ')
-			len--;
-		if (len == 0)
-			return 0;
-	}
-
-	while (i < len) {
-		if (s[i] == ' ') {
-			actual_len++;
-			while (i < len && s[i] == ' ')
-				i++;
-		} else {
-			actual_len++;
-			i++;
-		}
-	}
-
-	if (remove_trailing_space && s[i - 1] == ' ')
-		actual_len--;
-
-	return actual_len;
-}
-
-static void
-copy_collapsing_spaces (gchar *dst,
-			const gchar *src,
-			guint actual_len)
-{
-	const gchar *sp;
-	gchar *dp;
-	guint count;
-
-	if (actual_len == 0)
-		return;
-
-	sp = src;
-	dp = dst;
-	count = 0;
-
-	while (count < actual_len) {
-		if (*sp == ' ') {
-			*dp = ' ';
-			dp++, count++, sp++;
-			if (count < actual_len) {
-				while (*sp == ' ')
-					sp++;
-			}
-		} else {
-			*dp = *sp;
-			dp++, count++, sp++;
-		}
-	}
-}
-
 static void
 get_tags (const HTMLText *text,
 	  gchar *opening_tags,
@@ -167,21 +85,44 @@ get_tags (const HTMLText *text,
 	*ending_p = 0;
 }
 
+static void
+copy_helper (HTMLText *src,
+	     HTMLText *dest,
+	     guint offset,
+	     gint len)
+{
+	if (len < 0)
+		len = strlen (src->text);
+
+	dest->text = g_strndup (src->text + offset, len);
+	dest->text_len = len;
+
+	dest->font_style = src->font_style;
+	dest->color = src->color;
+	dest->color_allocated = src->color_allocated;
+}
+
 
 /* HTMLObject methods.  */
 
+static void
+reset (HTMLObject *self)
+{
+	HTMLText *text;
+
+	text = HTML_TEXT (self);
+	text->color_allocated = FALSE;
+
+	(* HTML_OBJECT_CLASS (parent_class)->reset) (self);
+}
+
 static void
 copy (HTMLObject *self,
       HTMLObject *dest)
 {
 	(* HTML_OBJECT_CLASS (parent_class)->copy) (self, dest);
 
-	HTML_TEXT (dest)->text = g_strdup (HTML_TEXT (self)->text);
-	HTML_TEXT (dest)->text_len = HTML_TEXT (self)->text_len;
-
-	HTML_TEXT (dest)->font_style = HTML_TEXT (self)->font_style;
-	HTML_TEXT (dest)->color = HTML_TEXT (self)->color;
-	HTML_TEXT (dest)->color_allocated = HTML_TEXT (self)->color_allocated;
+	copy_helper (HTML_TEXT (self), HTML_TEXT (dest), 0, -1);
 }
 
 static gboolean
@@ -289,6 +230,62 @@ queue_draw (HTMLText *text,
 	html_engine_queue_draw (engine, HTML_OBJECT (text));
 }
 
+static HTMLText *
+extract_text (HTMLText *text,
+	      guint offset,
+	      gint len)
+{
+	HTMLText *new;
+
+	if (len < 0)
+		len = text->text_len;
+
+	if (offset + len > text->text_len)
+		len = text->text_len - offset;
+
+	new = g_malloc (HTML_OBJECT (text)->klass->object_size);
+
+	(* HTML_OBJECT_CLASS (parent_class)->copy) (HTML_OBJECT (text), HTML_OBJECT (new));
+
+	copy_helper (HTML_TEXT (text), HTML_TEXT (new), offset, len);
+
+	return new;
+}
+
+/* #define DEBUG_NBSP */
+
+static void
+convert_nbsp (guchar *s, guint len)
+{
+	/* state of automata:
+	   0..Text
+	   1..Sequence <space>&nbsp;...&nbsp;
+	*/
+	guint state=0;
+
+#ifdef DEBUG_NBSP
+	printf ("convert_nbsp: ");
+#endif
+	while (len) {
+		if (*s == ENTITY_NBSP || *s == ' ') {
+			*s = state ? ENTITY_NBSP : ' ';
+			state = 1;
+		} else
+			state = 0;
+#ifdef DEBUG_NBSP
+		if (*s == ENTITY_NBSP)
+			printf ("&nbsp;");
+		else
+			printf ("%c", *s);
+#endif
+		len--;
+		s++;
+	}
+#ifdef DEBUG_NBSP
+	printf ("\n");
+#endif
+}
+
 static guint
 insert_text (HTMLText *text,
 	     HTMLEngine *engine,
@@ -296,16 +293,9 @@ insert_text (HTMLText *text,
 	     const gchar *s,
 	     guint len)
 {
-	gboolean remove_leading_space, remove_trailing_space;
-	const gchar *start;
 	gchar *new_buffer;
 	guint old_len;
 	guint new_len;
-	guint actual_len;
-
-	/* The following code is very stupid and quite inefficient, but it is
-           just for interactive editing so most likely people won't even
-           notice.  */
 
 	old_len = text->text_len;
 	if (offset > old_len) {
@@ -316,43 +306,25 @@ insert_text (HTMLText *text,
 		offset = old_len;
 	}
 
-	if (offset > 0 && text->text[offset - 1] == ' ')
-		remove_leading_space = TRUE;
-	else
-		remove_leading_space = FALSE;
-
-	if (text->text[offset] == ' ')
-		remove_trailing_space = TRUE;
-	else
-		remove_trailing_space = FALSE;
-
-	actual_len = calc_actual_len (s, len,
-				      remove_leading_space,
-				      remove_trailing_space,
-				      &start);
-	if (actual_len == 0)
-		return 0;
-
-	new_len = old_len + actual_len;
+	new_len    = old_len + len;
 	new_buffer = g_malloc (new_len + 1);
 
-	if (offset > 0)
-		memcpy (new_buffer, text->text, offset);
-
-	copy_collapsing_spaces (new_buffer + offset, start, actual_len);
-
-	if (offset < old_len)
-		memcpy (new_buffer + offset + actual_len,
-			text->text + offset,
-			old_len - offset);
-
+	/* concatenate strings */
+	memcpy (new_buffer,                text->text,        offset);
+	memcpy (new_buffer + offset,       s,                 len);
+	memcpy (new_buffer + offset + len, text->text+offset, old_len - offset);
 	new_buffer[new_len] = '\0';
 
-	g_free (text->text);
+	/* do <space>&nbsp;...&nbsp; hack */
+	convert_nbsp (new_buffer, new_len);
 
+	/* set new values */
+	g_free (text->text);
 	text->text = new_buffer;
 	text->text_len = new_len;
 
+	/* update */
+	html_object_change_set (HTML_OBJECT (text), HTML_CHANGE_ALL);
 	if (HTML_OBJECT (text)->parent != NULL) {
 		if (! html_object_relayout (HTML_OBJECT (text)->parent,
 					    engine,
@@ -360,7 +332,7 @@ insert_text (HTMLText *text,
 			html_engine_queue_draw (engine, HTML_OBJECT (text)->parent);
 	}
 
-	return actual_len;
+	return len;
 }
 
 static guint
@@ -389,18 +361,21 @@ remove_text (HTMLText *text,
 
 	new_len = old_len - len;
 
+	/* concat strings */
 	new_buffer = g_malloc (new_len + 1);
-
-	if (offset > 0)
-		memcpy (new_buffer, text->text, offset);
-
+	memcpy (new_buffer,          text->text,                offset);
 	memcpy (new_buffer + offset, text->text + offset + len, old_len - offset - len + 1);
 
-	g_free (text->text);
+	/* do <space>&nbsp;...&nbsp; hack */
+	convert_nbsp (new_buffer, new_len);
 
+	/* set new values */
+	g_free (text->text);
 	text->text = new_buffer;
 	text->text_len = new_len;
 
+	/* update */
+	html_object_change_set (HTML_OBJECT (text), HTML_CHANGE_ALL);
 	html_object_relayout (HTML_OBJECT (text)->parent, engine, HTML_OBJECT (text));
 	html_engine_queue_draw (engine, HTML_OBJECT (text)->parent);
 
@@ -445,20 +420,15 @@ split (HTMLText *self,
        guint offset)
 {
 	HTMLText *new;
-	guint len;
-	gchar *s;
-
-	len = strlen (self->text);
-
-	s = g_strdup (self->text + offset);
 
 	new = g_malloc (HTML_OBJECT (self)->klass->object_size);
 	html_text_init (new, HTML_TEXT_CLASS (HTML_OBJECT (self)->klass),
-			s, self->font_style, &self->color);
+			self->text + offset, -1, self->font_style, &self->color);
 
 	self->text = g_realloc (self->text, offset + 1);
 	self->text[offset] = '\0';
 	self->text_len = offset;
+	html_object_change_set (HTML_OBJECT (self), HTML_CHANGE_MIN_WIDTH);
 
 	return new;
 }
@@ -550,6 +520,13 @@ set_color (HTMLText *text,
 	}
 }
 
+static void
+destroy (HTMLObject *obj)
+{
+	g_free (HTML_TEXT (obj)->text);
+	HTML_OBJECT_CLASS (parent_class)->destroy (obj);
+}
+
 
 void
 html_text_type_init (void)
@@ -568,8 +545,8 @@ html_text_class_init (HTMLTextClass *klass,
 
 	html_object_class_init (object_class, type, object_size);
 
-	/* FIXME destroy */
-
+	object_class->destroy = destroy;
+	object_class->reset = reset;
 	object_class->copy = copy;
 	object_class->draw = draw;
 	object_class->accepts_cursor = accepts_cursor;
@@ -580,9 +557,10 @@ html_text_class_init (HTMLTextClass *klass,
 
 	/* HTMLText methods.  */
 
+	klass->queue_draw = queue_draw;
+	klass->extract_text = extract_text;
 	klass->insert_text = insert_text;
 	klass->remove_text = remove_text;
-	klass->queue_draw = queue_draw;
 	klass->split = split;
 	klass->get_font_style = get_font_style;
 	klass->get_color = get_color;
@@ -597,7 +575,8 @@ html_text_class_init (HTMLTextClass *klass,
 void
 html_text_init (HTMLText *text_object,
 		HTMLTextClass *klass,
-		gchar *text,
+		const gchar *text,
+		gint len,
 		GtkHTMLFontStyle font_style,
 		const GdkColor *color)
 {
@@ -607,9 +586,18 @@ html_text_init (HTMLText *text_object,
 
 	html_object_init (object, HTML_OBJECT_CLASS (klass));
 
-	text_object->text = text;
+	if (len == -1) {
+		text_object->text_len = strlen (text);
+		text_object->text = g_strdup (text);
+	} else {
+		text_object->text_len = len;
+		text_object->text = g_strndup (text, len);
+	}
+
+	/* do <space>&nbsp;...&nbsp; hack */
+	convert_nbsp (text_object->text, text_object->text_len);
+
 	text_object->font_style = font_style;
-	text_object->text_len = strlen (text);
 
 	if (color != NULL) {
 		text_object->color = *color;
@@ -623,19 +611,28 @@ html_text_init (HTMLText *text_object,
 }
 
 HTMLObject *
-html_text_new (gchar *text,
-	       GtkHTMLFontStyle font,
-	       const GdkColor *color)
+html_text_new_with_len (const gchar *text,
+			gint len,
+			GtkHTMLFontStyle font,
+			const GdkColor *color)
 {
 	HTMLText *text_object;
 
 	text_object = g_new (HTMLText, 1);
 
-	html_text_init (text_object, &html_text_class, text, font, color);
+	html_text_init (text_object, &html_text_class, text, len, font, color);
 
 	return HTML_OBJECT (text_object);
 }
 
+HTMLObject *
+html_text_new (const gchar *text,
+	       GtkHTMLFontStyle font,
+	       const GdkColor *color)
+{
+	return html_text_new_with_len (text, -1, font, color);
+}
+
 guint
 html_text_insert_text (HTMLText *text,
 		       HTMLEngine *engine,
@@ -677,6 +674,17 @@ html_text_queue_draw (HTMLText *text,
 	(* HT_CLASS (text)->queue_draw) (text, engine, offset, len);
 }
 
+HTMLText *
+html_text_extract_text (HTMLText *text,
+			guint offset,
+			gint len)
+{
+	g_return_val_if_fail (text != NULL, NULL);
+	g_return_val_if_fail (offset <= text->text_len, NULL);
+
+	return (* HT_CLASS (text)->extract_text) (text, offset, len);
+}
+
 HTMLText *
 html_text_split (HTMLText *text,
 		 guint offset)
@@ -738,12 +746,21 @@ html_text_set_font_style (HTMLText *text,
 	return (* HT_CLASS (text)->set_font_style) (text, engine, style);
 }
 
-void  html_text_set_color (HTMLText *text,
-			   HTMLEngine *engine,
-			   const GdkColor *color)
+void
+html_text_set_color (HTMLText *text,
+		     HTMLEngine *engine,
+		     const GdkColor *color)
 {
 	g_return_if_fail (text != NULL);
 	g_return_if_fail (color != NULL);
 
 	return (* HT_CLASS (text)->set_color) (text, engine, color);
 }
+
+void
+html_text_set_text (HTMLText *text, const gchar *new_text)
+{
+	g_free (text->text);
+	text->text = g_strdup (new_text);
+	html_object_change_set (HTML_OBJECT (text), HTML_CHANGE_ALL);
+}
