@@ -46,19 +46,23 @@ struct _HTMLText {
 struct _HTMLTextClass {
 	HTMLObjectClass object_class;
 
-	guint (* insert_text) (HTMLText *text, HTMLEngine *engine, guint offset, const gchar *p, guint len);
-	guint (* remove_text) (HTMLText *text, HTMLEngine *engine, guint offset, guint len);
-	void  (* queue_draw) (HTMLText *text, HTMLEngine *engine, guint offset, guint len);
+        guint        	   (* insert_text)    (HTMLText *text, HTMLEngine *engine,
+					       guint offset, const gchar *p, guint len);
+        guint        	   (* remove_text)    (HTMLText *text, HTMLEngine *engine,
+					       guint offset, guint len);
+        void         	   (* queue_draw)     (HTMLText *text, HTMLEngine *engine,
+					       guint offset, guint len);
+        	      
+        HTMLText     	 * (* split)          (HTMLText *text, guint offset);
+        HTMLText     	 * (* extract_text)   (HTMLText *text, guint offset, gint len);
+        void         	   (* merge)          (HTMLText *text, HTMLText *other, gboolean prepend);
+        gboolean     	   (* check_merge)    (HTMLText *self, HTMLText *text);
 
-	HTMLText *(* split) (HTMLText *text, guint offset);
-	void (* merge) (HTMLText *text, HTMLText *other, gboolean prepend);
-	gboolean (* check_merge) (HTMLText *self, HTMLText *text);
-	
-	GtkHTMLFontStyle (* get_font_style) (const HTMLText *text);
-	const GdkColor * (* get_color) (HTMLText *text, HTMLPainter *painter);
+	GtkHTMLFontStyle   (* get_font_style) (const HTMLText *text);
+	const GdkColor   * (* get_color)      (HTMLText *text, HTMLPainter *painter);
 
-	void (* set_font_style) (HTMLText *text, HTMLEngine *engine, GtkHTMLFontStyle style);
-	void (* set_color) (HTMLText *text, HTMLEngine *engine, const GdkColor *color);
+	void               (* set_font_style) (HTMLText *text, HTMLEngine *engine, GtkHTMLFontStyle style);
+	void               (* set_color)      (HTMLText *text, HTMLEngine *engine, const GdkColor *color);
 };
 
 
@@ -71,25 +75,35 @@ void        html_text_class_init   (HTMLTextClass    *klass,
 				    guint             object_size);
 void        html_text_init         (HTMLText         *text_object,
 				    HTMLTextClass    *klass,
-				    gchar            *text,
+				    const gchar      *text,
+				    gint              len,
 				    GtkHTMLFontStyle  font_style,
 				    const GdkColor   *color);
-HTMLObject *html_text_new          (gchar            *text,
+HTMLObject *html_text_new          (const gchar      *text,
 				    GtkHTMLFontStyle  font_style,
 				    const GdkColor   *color);
-guint       html_text_insert_text  (HTMLText         *text,
-				    HTMLEngine       *engine,
-				    guint             offset,
-				    const gchar      *p,
-				    guint             len);
-guint       html_text_remove_text  (HTMLText         *text,
-				    HTMLEngine       *engine,
-				    guint             offset,
-				    guint             len);
-void        html_text_queue_draw   (HTMLText         *text,
-				    HTMLEngine       *engine,
-				    guint             offset,
-				    guint             len);
+HTMLObject *html_text_new_with_len (const gchar      *text,
+				    gint              len,
+				    GtkHTMLFontStyle  font_style,
+				    const GdkColor   *color);
+
+guint     html_text_insert_text   (HTMLText    *text,
+				   HTMLEngine  *engine,
+				   guint        offset,
+				   const gchar *p,
+				   guint        len);
+HTMLText *html_text_extract_text  (HTMLText    *text,
+				   guint        offset,
+				   gint         len);
+guint     html_text_remove_text   (HTMLText    *text,
+				   HTMLEngine  *engine,
+				   guint        offset,
+				   guint        len);
+
+void  html_text_queue_draw  (HTMLText   *text,
+			     HTMLEngine *engine,
+			     guint       offset,
+			     guint       len);
 
 HTMLText *html_text_split        (HTMLText  *text,
 				  guint      offset);
@@ -103,11 +117,13 @@ GtkHTMLFontStyle  html_text_get_font_style  (const HTMLText *text);
 const GdkColor   *html_text_get_color       (HTMLText       *text,
 					     HTMLPainter    *painter);
 
-void  html_text_set_font_style  (HTMLText         *text,
-				 HTMLEngine       *engine,
-				 GtkHTMLFontStyle  style);
-void  html_text_set_color       (HTMLText         *text,
-				 HTMLEngine       *engine,
-				 const GdkColor   *color);
+void  html_text_set_font_style  (HTMLText             *text,
+				 HTMLEngine           *engine,
+				 GtkHTMLFontStyle      style);
+void  html_text_set_color       (HTMLText             *text,
+				 HTMLEngine           *engine,
+				 const GdkColor       *color);
+void  html_text_set_text        (HTMLText             *text,
+				 const gchar          *new_text);
 
 #endif /* _HTMLTEXT_H_ */
