@@ -32,6 +32,15 @@
 HTMLTextMasterClass html_text_master_class;
 static HTMLTextClass *parent_class = NULL;
 
+
+static void
+copy_text_master_members (HTMLTextMaster *self,
+			  HTMLTextMaster *dest)
+{
+	dest->select_start = self->select_start;
+	dest->select_length = self->select_length;
+}
+
 
 /* HTMLObject methods.  */
 
@@ -41,8 +50,8 @@ copy (HTMLObject *self,
 {
 	(* HTML_OBJECT_CLASS (parent_class)->copy) (self, dest);
 
-	HTML_TEXT_MASTER (dest)->select_start = HTML_TEXT_MASTER (self)->select_start;
-	HTML_TEXT_MASTER (dest)->select_length = HTML_TEXT_MASTER (self)->select_length;
+	copy_text_master_members (HTML_TEXT_MASTER (self),
+				  HTML_TEXT_MASTER (dest));
 }
 
 static void
@@ -62,9 +71,8 @@ calc_min_width (HTMLObject *self,
 	HTMLTextMaster *master;
 	GtkHTMLFontStyle font_style;
 	HTMLText *text;
-	const gchar *p;
-	gint min_width, run_width;
-	guint space_width;
+	gchar *p, *bp;
+	gint min_width, tmp_width;
 
 	master = HTML_TEXT_MASTER (self);
 	text = HTML_TEXT (self);
@@ -79,29 +87,27 @@ calc_min_width (HTMLObject *self,
 	}
 
 	min_width = 0;
-	run_width = 0;
-	p = text->text;
+	tmp_width = 0;
+	bp = p = text->text;
 
-	space_width = html_painter_calc_text_width (painter, " ", 1, font_style);
-
-	while (1) {
-		if (*p != ' ' && *p != 0) {
-			run_width += html_painter_calc_text_width (painter, p, 1, font_style);
-		} else {
-			if (run_width > min_width)
-				min_width = run_width;
-			run_width = 0;
-			if (*p == 0)
-				break;
+	while (bp && *p) {
+		bp = strchr (p, ' ');
+		if (bp) {
+			*bp = 0;
+		}
+		tmp_width = html_painter_calc_text_width (painter, p, (bp) ? bp-p : strlen (p), font_style);
+		if (tmp_width > min_width) {
+			min_width = tmp_width;
+		}
+		if (bp) {
+			*bp = ' ';
+			p = bp+1;
+			/* find the beggining of the next word */
+			while (*p == ' ')
+				p++;
 		}
-		p++;
 	}
 
-	/* printf ("T \"%s\" min width %d\n", min_width);
-	if (!self->next)
-		min_width -= space_width;
-		printf ("T \"%s\" min width %d\n", min_width); */
-
 	return min_width;
 }
 
@@ -233,8 +239,8 @@ select_range (HTMLObject *self,
 
 	master = HTML_TEXT_MASTER (self);
 
-	if (length < 0)
-		length = HTML_TEXT (self)->text_len;
+	if (length < 0 || length + offset > HTML_TEXT (self)->text_len)
+		length = HTML_TEXT (self)->text_len - offset;
 
 	if (offset != master->select_start || length != master->select_length)
 		changed = TRUE;
@@ -259,8 +265,7 @@ select_range (HTMLObject *self,
 			else
 				was_selected = FALSE;
 
-			if (offset + length > slave->posStart &&
-			    offset < max)
+			if (offset + length > slave->posStart && offset < max)
 				is_selected = TRUE;
 			else
 				is_selected = FALSE;
@@ -306,7 +311,6 @@ get_selection (HTMLObject *self,
 	HTMLObject *new;
 	HTMLText *text_self;
 	guint select_start, select_length;
-	gchar *text;
 
 	if (! self->selected)
 		return NULL;
@@ -316,9 +320,8 @@ get_selection (HTMLObject *self,
 
 	text_self = HTML_TEXT (self);
 
-	text = g_strndup (text_self->text + select_start, select_length);
-
-	new = html_text_master_new (text, text_self->font_style, & text_self->color);
+	new = html_text_master_new_with_len (text_self->text + select_start,
+					     select_length, text_self->font_style, & text_self->color);
 
 	if (size_return != NULL)
 		*size_return = select_length;
@@ -326,6 +329,28 @@ get_selection (HTMLObject *self,
 	return new;
 }
 
+static void
+append_selection_string (HTMLObject *self,
+			 GString *buffer)
+{
+	HTMLTextMaster *text_master;
+	const gchar *p;
+	guint i;
+
+	text_master = HTML_TEXT_MASTER (self);
+	if (text_master->select_length == 0)
+		return;
+
+	/* FIXME: we need a `g_string_append()' that takes the number of
+           characters to append as an extra parameter.  */
+
+	p = HTML_TEXT (text_master)->text + text_master->select_start;
+	for (i = 0; i < text_master->select_length; i++) {
+		g_string_append_c (buffer, *p);
+		p++;
+	}
+}
+
 static void
 get_cursor (HTMLObject *self,
 	    HTMLPainter *painter,
@@ -418,6 +443,21 @@ queue_draw (HTMLText *text,
 	}
 }
 
+static HTMLText *
+extract_text (HTMLText *text,
+	      guint offset,
+	      gint len)
+{
+	HTMLText *new;
+
+	new = (* HTML_TEXT_CLASS (parent_class)->extract_text) (text, offset, len);
+
+	copy_text_master_members (HTML_TEXT_MASTER (text),
+				  HTML_TEXT_MASTER (new));
+
+	return new;
+}
+
 static HTMLText *
 split (HTMLText *self,
        guint offset)
@@ -465,6 +505,32 @@ split (HTMLText *self,
 	return new;
 }
 
+static void
+calculate_new_selection (HTMLText *obj1,
+			 HTMLText *obj2,
+			 guint *select_start_return,
+			 guint *select_length_return)
+{
+	if (HTML_TEXT_MASTER (obj1)->select_length == 0) {
+		*select_start_return = (HTML_TEXT_MASTER (obj2)->select_start
+					+ HTML_TEXT (obj1)->text_len);
+		*select_length_return = HTML_TEXT_MASTER (obj2)->select_length;
+		return;
+	}
+
+	if (HTML_TEXT_MASTER (obj2)->select_length == 0) {
+		*select_start_return = HTML_TEXT_MASTER (obj1)->select_start;
+		*select_length_return = HTML_TEXT_MASTER (obj1)->select_length;
+		return;
+	}
+
+	*select_start_return = HTML_TEXT_MASTER (obj1)->select_start;
+	*select_length_return = (HTML_TEXT (obj1)->text_len
+				 - HTML_TEXT_MASTER (obj1)->select_start
+				 + HTML_TEXT_MASTER (obj2)->select_start
+				 + HTML_TEXT_MASTER (obj2)->select_length);
+}
+
 static void
 merge (HTMLText *self,
        HTMLText *other,
@@ -477,38 +543,10 @@ merge (HTMLText *self,
 	if (other->text_len == 0)
 		return;
 
-	/* Calculate the new selection.  */
-
-	if (prepend) {
-		if (HTML_TEXT_MASTER (self)->select_length == 0) {
-			select_start = HTML_TEXT_MASTER (other)->select_start;
-			select_length = HTML_TEXT_MASTER (other)->select_length;
-		} else if (HTML_TEXT_MASTER (other)->select_length == 0) {
-			select_start = HTML_TEXT (other)->text_len;
-			select_length = HTML_TEXT_MASTER (self)->select_length;
-		} else {
-			select_start = HTML_TEXT_MASTER (other)->select_start;
-			select_length = (HTML_TEXT (other)->text_len
-					 + HTML_TEXT_MASTER (self)->select_start
-					 + HTML_TEXT_MASTER (self)->select_length
-					 - HTML_TEXT_MASTER (other)->select_start);
-		}
-	} else {
-		if (HTML_TEXT_MASTER (self)->select_length == 0) {
-			select_start = (HTML_TEXT_MASTER (other)->select_length
-					+ HTML_TEXT (self)->text_len);
-			select_length = HTML_TEXT (other)->text_len;
-		} else if (HTML_TEXT_MASTER (other)->select_length == 0) {
-			select_start = HTML_TEXT_MASTER (self)->select_start;
-			select_length = HTML_TEXT_MASTER (self)->select_length;
-		} else {
-			select_start = HTML_TEXT_MASTER (self)->select_start;
-			select_length = (HTML_TEXT (self)->text_len
-					 + HTML_TEXT_MASTER (other)->select_start
-					 + HTML_TEXT_MASTER (other)->select_length
-					 - HTML_TEXT_MASTER (self)->select_start);
-		}
-	}
+	if (prepend)
+		calculate_new_selection (other, self, &select_start, &select_length);
+	else
+		calculate_new_selection (self, other, &select_start, &select_length);
 
 	total_length = self->text_len + other->text_len;
 	new_text = g_malloc (total_length + 1);
@@ -616,10 +654,12 @@ html_text_master_class_init (HTMLTextMasterClass *klass,
 	object_class->check_point = check_point;
 	object_class->select_range = select_range;
 	object_class->get_selection = get_selection;
+	object_class->append_selection_string = append_selection_string;
 
 	/* HTMLText methods.  */
 
 	text_class->queue_draw = queue_draw;
+	text_class->extract_text = extract_text;
 	text_class->split = split;
 	text_class->merge = merge;
 	text_class->insert_text = insert_text;
@@ -631,7 +671,8 @@ html_text_master_class_init (HTMLTextMasterClass *klass,
 void
 html_text_master_init (HTMLTextMaster *master,
 		       HTMLTextMasterClass *klass,
-		       gchar *text,
+		       const gchar *text,
+		       gint len,
 		       GtkHTMLFontStyle font_style,
 		       const GdkColor *color)
 {
@@ -641,25 +682,34 @@ html_text_master_init (HTMLTextMaster *master,
 	html_text = HTML_TEXT (master);
 	object = HTML_OBJECT (master);
 
-	html_text_init (html_text, HTML_TEXT_CLASS (klass), text, font_style, color);
+	html_text_init (html_text, HTML_TEXT_CLASS (klass), text, len, font_style, color);
 
 	master->select_start = 0;
 	master->select_length = 0;
 }
 
 HTMLObject *
-html_text_master_new (gchar *text,
-		      GtkHTMLFontStyle font_style,
-		      const GdkColor *color)
+html_text_master_new_with_len (const gchar *text,
+			       gint len,
+			       GtkHTMLFontStyle font_style,
+			       const GdkColor *color)
 {
 	HTMLTextMaster *master;
 
 	master = g_new (HTMLTextMaster, 1);
-	html_text_master_init (master, &html_text_master_class, text, font_style, color);
+	html_text_master_init (master, &html_text_master_class, text, len, font_style, color);
 
 	return HTML_OBJECT (master);
 }
 
+HTMLObject *
+html_text_master_new (const gchar *text,
+		      GtkHTMLFontStyle font_style,
+		      const GdkColor *color)
+{
+	return html_text_master_new_with_len (text, -1, font_style, color);
+}
+
 
 HTMLObject *
 html_text_master_get_slave (HTMLTextMaster *master, guint offset)
@@ -683,7 +733,7 @@ html_text_master_trail_space_width (HTMLTextMaster *master, HTMLPainter *painter
 {
 	HTMLText *text = HTML_TEXT (master);
 
-	if (text->text [text->text_len-1] == ' ') {
+	if (text->text_len > 0 && text->text [text->text_len-1] == ' ') {
 		GtkHTMLFontStyle font_style;
 
 		font_style = html_text_get_font_style (text);
