@@ -172,17 +172,6 @@ fit_line (HTMLObject *o,
 
 	text += textslave->posStart;
 
-	/* FIXME does it really have to be this sick? and should we also eat trailing spaces */
-	/* eat up spaces at the begining of lines */
-
-	if (startOfLine) {
-		while (*text == ' ') {
-			text++;
-			textslave->posLen--;
-			textslave->posStart++;
-		}		
-	}
-	
 	o->width = html_painter_calc_text_width (painter, text, textslave->posLen, font_style);
 	if (o->width <= widthLeft || textslave->posLen <= 1 || widthLeft < 0) {
 		/* Text fits completely */
@@ -487,6 +476,10 @@ html_text_slave_init (HTMLTextSlave *slave,
 	slave->posStart = posStart;
 	slave->posLen = posLen;
 	slave->owner = owner;
+
+	/* text slaves has always min_width 0 */
+	object->min_width = 0;
+	object->change   &= ~HTML_CHANGE_MIN_WIDTH;
 }
 
 HTMLObject *
@@ -537,6 +530,9 @@ html_text_slave_get_offset_for_pointer (HTMLTextSlave *slave,
 gint
 html_text_slave_trail_space_width (HTMLTextSlave *slave, HTMLPainter *painter)
 {
+	if (slave->posLen == 0)
+		return 0;
+
 	if (HTML_TEXT (slave->owner)->text [slave->posStart + slave->posLen - 1] == ' ') {
 		GtkHTMLFontStyle font_style;
 
