@@ -16,6 +16,8 @@
     Boston, MA 02111-1307, USA.
 */
 
+#define MEMDEBUG
+
 #include <gnome.h>
 #include <gtk/gtk.h>
 
@@ -231,12 +233,6 @@ create_toolbars (GtkWidget *app)
 				 home_cb, NULL);
 	gtk_toolbar_append_space (GTK_TOOLBAR (toolbar));
 
-	gtk_toolbar_append_item (GTK_TOOLBAR (toolbar),
-				 NULL, 
-				 "Search bar",
-				 "Search",
-				 gnome_stock_new_with_icon (GNOME_STOCK_PIXMAP_SEARCH),
-				 NULL, NULL);
 	animator = gnome_animator_new_with_size (32, 32);
 
 	if (g_file_exists("32.png"))
@@ -437,12 +433,9 @@ reload_cb (GtkWidget *widget, gpointer data)
 static void
 stop_cb (GtkWidget *widget, gpointer data)
 {
-	if (html_stream_handle != NULL) {
-		gtk_html_end (html, html_stream_handle, GTK_HTML_STREAM_OK);
-		html_stream_handle = NULL;
-	}
 	/* Kill all requests */
 	HTNet_killAll();
+	html_stream_handle = NULL;
 }
 
 static void
@@ -459,8 +452,6 @@ static int
 on_button_press_event (GtkWidget *widget, GdkEventButton *event)
 {
 	GtkMenu *menu;
-	GdkEventButton *event_button;
-	int value;
 	
 	g_return_val_if_fail (widget != NULL, FALSE);
 	g_return_val_if_fail (event != NULL, FALSE);
@@ -471,34 +462,10 @@ on_button_press_event (GtkWidget *widget, GdkEventButton *event)
 	menu = GTK_MENU (popup_menu);
 	
 	if (event->type == GDK_BUTTON_PRESS) {
-		event_button = (GdkEventButton *) event;
-		if (event_button->button == 3) {
+
+		if (event->button == 3) {
 			gtk_menu_popup (menu, NULL, NULL, NULL, NULL, 
-					event_button->button, event_button->time);
-			return TRUE;
-		}
-		/* Mouse wheel scroll up */
-		if (event_button->button == 4) {
-
-			value = GTK_LAYOUT (html)->vadjustment->value - GTK_LAYOUT (html)->vadjustment->step_increment * 3;
-
-			if(value < GTK_LAYOUT (html)->vadjustment->lower)
-				value = GTK_LAYOUT (html)->vadjustment->lower;
-
-			gtk_adjustment_set_value (GTK_LAYOUT (html)->vadjustment, value);
-				
-			return TRUE;
-		} 
-		/* Mouse wheel scroll down */
-		if (event_button->button == 5) {
-
-			value = GTK_LAYOUT (html)->vadjustment->value + GTK_LAYOUT (html)->vadjustment->step_increment * 3;
-
-			if(value > (GTK_LAYOUT (html)->vadjustment->upper - GTK_LAYOUT (html)->vadjustment->page_size))
-				value = GTK_LAYOUT (html)->vadjustment->upper - GTK_LAYOUT (html)->vadjustment->page_size;
-
-			gtk_adjustment_set_value (GTK_LAYOUT (html)->vadjustment, value);
-
+					event->button, event->time);
 			return TRUE;
 		}
 	}
@@ -744,6 +711,7 @@ parse_href (const gchar *s)
 				/* FIXME?  This is a bit sucky.  */
 				t = g_strconcat (html_url_get_protocol (baseURL),
 						 ":", s, NULL);
+				html_url_destroy (tmpurl);
 				tmpurl = html_url_new (t);
 				retval = html_url_to_string (tmpurl);
 				html_url_destroy (tmpurl);
@@ -751,6 +719,7 @@ parse_href (const gchar *s)
 			} else {
 				/* Single slash at the beginning.  */
 
+				html_url_destroy (tmpurl);
 				tmpurl = html_url_dup (baseURL,
 						       HTML_URL_DUP_NOPATH);
 				html_url_set_path (tmpurl, s);
@@ -758,12 +727,14 @@ parse_href (const gchar *s)
 				html_url_destroy (tmpurl);
 			}
 		} else {
+			html_url_destroy (tmpurl);
 			tmpurl = html_url_append_path (baseURL, s);
 			retval = html_url_to_string (tmpurl);
 			html_url_destroy (tmpurl);
 		}
 	} else {
 		retval = html_url_to_string (tmpurl);
+		html_url_destroy (tmpurl);
 	}
 
 	return retval;
@@ -1018,8 +989,12 @@ main (gint argc, gchar *argv[])
 	GtkWidget *app, *bar;
 	GtkWidget *html_widget;
 	GtkWidget *scrolled_window;
+	GdkColor   bgColor = {0, 0xdfff, 0xdfff, 0xffff};
 	poptContext ctx;
-	
+
+#ifdef MEMDEBUG
+	void *p = malloc (1024);	/* to make linker happy with ccmalloc */
+#endif
 	gnome_init_with_popt_table (PACKAGE, VERSION,
 				    argc, argv, options, 0, &ctx);
 	glibwww_init(PACKAGE, VERSION);
@@ -1058,6 +1033,8 @@ main (gint argc, gchar *argv[])
 	html_widget = gtk_html_new ();
 	html = GTK_HTML (html_widget);
 	gtk_html_load_empty (html);
+	gtk_html_set_default_background_color (GTK_HTML (html_widget), &bgColor);
+	/* gtk_html_set_editable (GTK_HTML (html_widget), TRUE); */
 	
 	gtk_container_add (GTK_CONTAINER (scrolled_window), html_widget);
 
@@ -1131,5 +1108,11 @@ main (gint argc, gchar *argv[])
 
 	gtk_main ();
 
+#ifdef MEMDEBUG
+
+	// gtk_widget_unref (html_widget);
+	free (p);
+#endif
+
 	return 0;
 }
